# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Application manifest for datetime.
"""

from django.utils.translation import gettext_lazy as _

backup = {
    'data': {
        'files': ['/etc/localtime']
    },
    'services': ['systemd-timedated'],
}

tags = [_('Network time'), _('Timezone')]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test the backups action script.
"""

import os
import pathlib
import subprocess
import uuid

import pytest

from plinth.modules import backups
from plinth.modules.backups import privileged
from plinth.modules.backups.repository import BorgRepository, SshBorgRepository
from plinth.tests import config as test_config

pytestmark = pytest.mark.usefixtures('needs_root', 'needs_borg', 'load_cfg',
                                     'mock_privileged')

privileged_modules_to_mock = ['plinth.modules.backups.privileged']

# try to access a non-existing url and a URL that exists but does not
# grant access
_dummy_credentials = {'ssh_password': 'invalid_password'}
_repokey_encryption_passphrase = '12345'


@pytest.fixture(name='needs_borg')
def fixture_needs_borg():
    """Return whether borg is installed on the system."""
    try:
        subprocess.run(['borg', '--version'], stdout=subprocess.DEVNULL,
                       stderr=subprocess.DEVNULL, check=True)
    except FileNotFoundError:
        pytest.skip('Needs borg installed')


@pytest.fixture(name='needs_ssh_config')
def fixture_needs_ssh_config():
    """Skip test if SSH details is not available in test configuration."""
    if not test_config.backups_ssh_path:
        pytest.skip('Needs SSH password provided')


@pytest.fixture(name='data_directory')
def fixture_data_directory():
    """Return directory where backup data is stored."""
    return pathlib.Path(__file__).parent / 'backup_data'


@pytest.fixture(name='backup_directory')
def fixture_backup_directory(tmp_path):
    """Create and cleanup a backup directory."""
    return tmp_path


def test_nonexisting_repository(backup_directory):
    """Test that non-existent directory as borg repository throws error."""
    nonexisting_dir = backup_directory / 'does_not_exist'
    repository = BorgRepository(str(nonexisting_dir))
    with pytest.raises(backups.errors.BorgRepositoryDoesNotExistError):
        repository.get_info()


def test_empty_dir(backup_directory):
    """Test that empty directory as borg repository throws error."""
    empty_dir = backup_directory / 'empty_dir'
    empty_dir.mkdir()
    repository = BorgRepository(str(empty_dir))
    with pytest.raises(backups.errors.BorgRepositoryDoesNotExistError):
        repository.get_info()


def test_create_unencrypted_repository(backup_directory):
    """Test creating an unencrypted repository."""
    path = backup_directory / 'borgbackup'
    repository = BorgRepository(str(path))
    repository.initialize()
    info = repository.get_info()
    assert 'encryption' in info


def test_create_export_delete_archive(data_directory, backup_directory):
    """
    - Create a repo
    - Create an archive
    - Verify archive content
    - Delete archive
    """
    repo_name = 'test_create_and_delete'
    archive_name = 'first_archive'
    archive_comment = 'test_archive_comment'
    path = backup_directory / repo_name

    repository = BorgRepository(str(path))
    repository.initialize()
    archive_path = "::".join([str(path), archive_name])
    privileged.create_archive(archive_path, [str(data_directory)],
                              archive_comment)

    archive = repository.list_archives()[0]
    assert archive['name'] == archive_name
    assert archive['comment'] == archive_comment

    repository.delete_archive(archive_name)
    content = repository.list_archives()
    assert not content


@pytest.mark.usefixtures('needs_ssh_config')
def test_remote_backup_actions():
    """
    Test creating an encrypted remote repository using borg directly.

    This relies on borgbackups being installed on the remote machine.
    """
    credentials = _get_credentials(add_encryption_passphrase=True)
    path = os.path.join(test_config.backups_ssh_path, str(uuid.uuid1()))
    privileged.init(path, 'repokey', **_get_borg_arguments(credentials))

    info = privileged.info(path, **_get_borg_arguments(credentials))
    assert info['encryption']['mode'] == 'repokey'


def _get_borg_arguments(credentials):
    """Get credential arguments for running borg privileged actions."""
    return {
        'passphrase': credentials.get('encryption_passphrase', None),
        'ssh_keyfile': credentials.get('ssh_keyfile')
    }


@pytest.mark.usefixtures('needs_ssh_config')
def test_sshfs_mount_password():
    """Test (un)mounting if password for a remote location is given"""
    credentials = _get_credentials()
    ssh_path = test_config.backups_ssh_path

    repository = SshBorgRepository(ssh_path, credentials)
    repository.mount()
    assert repository.is_mounted
    repository.umount()
    assert not repository.is_mounted


@pytest.mark.usefixtures('needs_ssh_config')
def test_sshfs_mount_keyfile():
    """Test (un)mounting if keyfile for a remote location is given"""
    credentials = _get_credentials()
    ssh_path = test_config.backups_ssh_path

    repository = SshBorgRepository(ssh_path, credentials)
    repository.mount()
    assert repository.is_mounted
    repository.umount()
    assert not repository.is_mounted


def test_access_nonexisting_url():
    """Test accessing a non-existent URL."""
    repo_url = "user@%s.com.au:~/repo" % str(uuid.uuid1())
    repository = SshBorgRepository(repo_url, _dummy_credentials)
    with pytest.raises(backups.errors.BorgRepositoryDoesNotExistError):
        repository.get_info()


def test_inaccessible_repo_url():
    """Test accessing an existing URL with wrong credentials."""
    repo_url = 'user@heise.de:~/repo'
    repository = SshBorgRepository(repo_url, _dummy_credentials)
    with pytest.raises(backups.errors.BorgError):
        repository.get_info()


def _get_credentials(add_encryption_passphrase=False):
    """
    Get access params for a remote location.
    Return an empty dict if no valid access params are found.
    """
    credentials = {}
    if test_config.backups_ssh_password:
        credentials['ssh_password'] = test_config.backups_ssh_password
    elif test_config.backups_ssh_keyfile:
        credentials['ssh_keyfile'] = test_config.backups_ssh_keyfile

    if add_encryption_passphrase:
        credentials['encryption_passphrase'] = \
            _repokey_encryption_passphrase

    return credentials



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Custom password hashers suitable for home servers.
"""

from django.contrib.auth.hashers import Argon2PasswordHasher


class Argon2PasswordHasherLowMemory(Argon2PasswordHasher):
    """Argon2 password hasher that uses less CPU and RAM than Django's default.

    Derive from and override the default complexity parameters for Django. In
    Django 2.2, the defaults are time: 2, memory: 512 and parallelism: 2. In
    Django 3.2, the defaults are time: 2, memory: 102400, parallelism: 8. This
    takes more than 3 seconds per verification on a Lime2 board.

    On a Pioneer Edition, Olimex Lime2 board, the selected parameters result in
    about 200ms for password verification:

    $ python3 -m argon2 -p 2 -m 4096
    Running Argon2id 100 times with:
    hash_len: 16 bytes
    memory_cost: 4096 KiB
    parallelism: 2 threads
    time_cost: 2 iterations

    Measuring...

    2.17e+02ms per password verification

    """
    time_cost = 2  # Iterations
    memory_cost = 4096  # KiB
    parallelism = 2  # Threads



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Django views for user app."""

import functools
import importlib.metadata
import json
import logging
import re
import uuid

import axes.utils
import django.views.generic
import fido2.cbor
import fido2.features
import fido2.utils
from django import shortcuts
from django.contrib import messages
from django.contrib.auth import login as auth_login
from django.contrib.auth import logout as auth_logout
from django.contrib.auth import update_session_auth_hash
from django.contrib.auth.models import User
from django.contrib.auth.views import LoginView as DjangoLoginView
from django.contrib.messages.views import SuccessMessageMixin
from django.core.exceptions import PermissionDenied
from django.db import IntegrityError
from django.http import HttpResponseRedirect, JsonResponse
from django.urls import reverse, reverse_lazy
from django.utils.decorators import method_decorator
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy
from django.views.decorators.http import require_POST
from django.views.generic.edit import (CreateView, DeleteView, FormView,
                                       UpdateView)
from fido2 import webauthn
from fido2.server import Fido2Server
from fido2.webauthn import AttestedCredentialData, AuthenticationResponse

import plinth.modules.ssh.privileged as ssh_privileged
from plinth import translation
from plinth.models import UserPasskey
from plinth.modules import first_boot
from plinth.utils import is_user_admin
from plinth.version import Version
from plinth.views import AppView, json_exception

from . import privileged
from .forms import (AuthenticationForm, CaptchaForm, CreateUserForm,
                    FirstBootForm, UserChangePasswordForm, UserUpdateForm)

# Enable fido2 basic features
if Version(importlib.metadata.version('fido2')) < Version('2.0.0'):
    fido2.features.webauthn_json_mapping.enabled = True  # type: ignore

logger = logging.getLogger(__name__)


class LoginView(DjangoLoginView):
    """View to login to FreedomBox and set language preference."""

    redirect_authenticated_user = True
    template_name = 'users_login.html'
    form_class = AuthenticationForm

    def dispatch(self, request, *args, **kwargs):
        """Handle a request and return a HTTP response."""
        response = super().dispatch(request, *args, **kwargs)
        if request.user.is_authenticated:
            translation.set_language(request, response,
                                     request.user.userprofile.language)

        return response


class CaptchaView(FormView):
    """A simple form view with a CAPTCHA image.

    When a user performs too many login attempts, they will no longer be able
    to login with the typical login view. They will be redirected to this view.
    On successfully solving the CAPTCHA in this form, their ability to use the
    login form will be reset.
    """

    template_name = 'users_captcha.html'
    form_class = CaptchaForm

    def form_valid(self, form):
        """Reset login attempts and redirect to login page."""
        axes.utils.reset_request(self.request)
        return shortcuts.redirect('users:login')


@require_POST
def logout(request):
    """Logout an authenticated user, remove SSO cookie and redirect to home."""
    auth_logout(request)
    response = shortcuts.redirect('index')

    # HACK: Remove Apache OpenID Connect module's session. This will logout all
    # the apps using mod_auth_openidc for their authentication and
    # authorization. A better way to do this is to implement OpenID Connect's
    # Back-Channel Logout[1] or using OpenID Connect Session Management[2].
    # With this scheme, each application will register a logout URL during
    # client registration. The OpenID Provider (FreedomBox service) will call
    # this URL with appropriate parameters to perform logout on all the apps.
    # Support for OpenID Connect Back-Channel Logout is currently under
    # review[3].
    # 1. https://openid.net/specs/openid-connect-backchannel-1_0.html
    # 2. https://openid.net/specs/openid-connect-session-1_0.html
    # 3. https://github.com/django-oauth/django-oauth-toolkit/pull/1573
    response.delete_cookie('mod_auth_openidc_session')

    messages.success(request, _('Logged out successfully.'))
    return response


class ContextMixin:
    """Mixin to add 'title' to the template context."""

    def get_context_data(self, **kwargs):
        """Add self.title to template context."""
        context = super().get_context_data(**kwargs)
        context['title'] = getattr(self, 'title', '')
        return context


class UserCreate(ContextMixin, SuccessMessageMixin, CreateView):
    """View to create a new user."""

    form_class = CreateUserForm
    template_name = 'users_create.html'
    model = User
    success_message = gettext_lazy('User %(username)s created.')
    success_url = reverse_lazy('users:create')
    title = gettext_lazy('Create User')

    def get_form_kwargs(self):
        """Make the request object available to the form."""
        kwargs = super().get_form_kwargs()
        kwargs['request'] = self.request
        return kwargs

    def get_success_url(self):
        """Return the URL to redirect to in case of successful updation."""
        return reverse('users:index')


class UserList(AppView, ContextMixin, django.views.generic.ListView):
    """View to list users."""

    model = User
    template_name = 'users_list.html'
    title = gettext_lazy('Users')
    app_id = 'users'


class UserUpdate(ContextMixin, SuccessMessageMixin, UpdateView):
    """View to update a user's details."""

    template_name = 'users_update.html'
    model = User
    form_class = UserUpdateForm
    slug_field = 'username'
    success_message = gettext_lazy('User %(username)s updated.')
    title = gettext_lazy('Edit User')

    def dispatch(self, request, *args, **kwargs):
        """Handle a request and return a HTTP response."""
        if self.request.user.get_username() != self.kwargs['slug'] \
                and not is_user_admin(self.request):
            raise PermissionDenied

        return super().dispatch(request, *args, **kwargs)

    def get_form_kwargs(self):
        """Make the request object available to the form."""
        kwargs = super().get_form_kwargs()
        kwargs['request'] = self.request
        kwargs['username'] = self.object.username
        return kwargs

    def get_initial(self):
        """Return the data for initial form load."""
        initial = super().get_initial()
        try:
            ssh_keys = ssh_privileged.get_keys(self.object.username)
            initial['ssh_keys'] = ssh_keys.strip()
            initial['language'] = self.object.userprofile.language
        except Exception:
            pass

        return initial

    def get_success_url(self):
        """Return the URL to redirect to in case of successful updation."""
        return reverse('users:edit', kwargs={'slug': self.object.username})

    def form_valid(self, form):
        """Set the user language if necessary."""

        is_user_deleted = form.cleaned_data.get('delete')
        if is_user_deleted:
            self.success_message = gettext_lazy('User %(username)s deleted.')
        response = super().form_valid(form)
        if is_user_deleted:
            return HttpResponseRedirect(reverse_lazy('users:index'))

        # If user is updating their own profile then set the language for
        # current session too.
        if self.object.username == self.request.user.username:
            translation.set_language(self.request, response,
                                     self.request.user.userprofile.language)

        return response


class UserChangePassword(ContextMixin, SuccessMessageMixin, FormView):
    """View to change user password."""

    template_name = 'users_change_password.html'
    form_class = UserChangePasswordForm
    title = gettext_lazy('Change Password')
    success_message = gettext_lazy('Password changed successfully.')

    def dispatch(self, request, *args, **kwargs):
        """Handle a request and return a HTTP response."""
        if self.request.user.get_username() != self.kwargs['slug'] \
                and not is_user_admin(self.request):
            raise PermissionDenied

        return super().dispatch(request, *args, **kwargs)

    def get_form_kwargs(self):
        """Make the user object available to the form."""
        kwargs = super().get_form_kwargs()
        kwargs['request'] = self.request
        kwargs['user'] = User.objects.get(username=self.kwargs['slug'])
        return kwargs

    def get_success_url(self):
        """Return the URL to go to on successful sumbission."""
        return reverse('users:edit', kwargs={'slug': self.kwargs['slug']})

    def form_valid(self, form):
        """Save the form if the submission is valid.

        Django user session authentication hashes are based on password to have
        the ability to logout all sessions on password change.  Update the
        session authentications to ensure that the current sessions is not
        logged out.
        """
        form.save()
        update_session_auth_hash(self.request, form.user)
        return super().form_valid(form)


class FirstBootView(django.views.generic.CreateView):
    """Create user account and log the user in."""

    template_name = 'users_firstboot.html'
    form_class = FirstBootForm

    def dispatch(self, request, *args, **kwargs):
        """Check if there is no possibility to create a new admin account."""
        if request.method == 'POST' and 'skip' in request.POST:
            first_boot.mark_step_done('users_firstboot')
            return HttpResponseRedirect(reverse(first_boot.next_step()))

        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, *args, **kwargs):
        """Add admin users to context data."""
        context = super().get_context_data(*args, **kwargs)
        context['admin_users'] = privileged.get_group_users('admin')
        return context

    def get_form_kwargs(self):
        """Make request available to the form (to insert messages)."""
        kwargs = super().get_form_kwargs()
        kwargs['request'] = self.request
        return kwargs

    def get_success_url(self):
        """Return the next first boot step after valid form submission."""
        return reverse(first_boot.next_step())


def get_fido2_server(domain: str) -> Fido2Server:
    """Return an instance of a Fido2Server."""
    relying_party = webauthn.PublicKeyCredentialRpEntity(
        id=domain, name='FreedomBox')
    return Fido2Server(relying_party)


def require_owner_or_admin(view_func):
    """Decorator to check if the view is called by owner or admin."""

    @functools.wraps(view_func)
    def _wrapped_view(request, *args, **kwargs):
        """Wrap a view method and check ownership is valid."""
        if (kwargs['username'] != request.user.get_username()
                and not is_user_admin(request)):
            raise PermissionDenied

        return view_func(request, *args, **kwargs)

    return _wrapped_view


@method_decorator(require_owner_or_admin, name='dispatch')
class PasskeysList(ContextMixin, django.views.generic.ListView):
    """View to show a list of current passkeys for the user."""
    model = UserPasskey
    template_name = 'users_passkeys.html'
    title = gettext_lazy('Passkeys')

    def get_queryset(self):
        """Show only list of passkeys for this user."""
        return self.model.objects.filter(
            user__username=self.kwargs['username']).order_by('created_time')


@json_exception
@require_owner_or_admin
@require_POST
def passkey_add_begin(request, username):
    """Verify passkey registration and add passkey to user's account."""
    user = User.objects.get(username=username)

    # Domain
    domain = request.get_host().partition(':')[0]

    # User
    username = user.get_username()
    user_entity = webauthn.PublicKeyCredentialUserEntity(
        id=user.id, name=username, display_name=username)

    # Begin registration
    server = get_fido2_server(domain)
    create_options, state = server.register_begin(
        user_entity,
        user_verification=webauthn.UserVerificationRequirement.REQUIRED,
        resident_key_requirement=webauthn.ResidentKeyRequirement.PREFERRED)

    logger.info('Passkey creation began for user - %s (%s)', username, user.id)

    request.session['fido2_server_state'] = state
    return JsonResponse(dict(create_options))


def _passkey_get_next_name(user):
    """Return a suitable name for the next passkey of the given user."""
    number = 1
    for passkey in user.passkeys.all():
        match = re.match(r'^Key (\d+)$', passkey.name)
        if match:
            number = max(number, int(match[1]) + 1)

    return f'Key {number}'


@json_exception
@require_owner_or_admin
@require_POST
def passkey_add_complete(request, username):
    """Verify passkey registration and add passkey to user's account."""

    def _response(result: bool, error_string: str):
        """Return a JsonResponse object."""
        status = 200
        if not result:
            status = 400
            logger.error('Error completing passkey registration: %s',
                         error_string)

        return JsonResponse({
            'result': result,
            'error_string': error_string
        }, status=status)

    user = User.objects.get(username=username)

    try:
        response = json.loads(request.body)
    except json.decoder.JSONDecodeError as exception:
        return _response(False, str(exception))

    # Domain
    domain = request.get_host().partition(':')[0]

    # State
    state = request.session.get('fido2_server_state')

    # Complete registration
    server = get_fido2_server(domain)
    try:
        authenticator_data = server.register_complete(state, response)
        logger.info('Passkey creation completed for user - %s (%s)',
                    user.get_username(), user.id)
        assert authenticator_data.is_user_present()
        assert authenticator_data.is_user_verified()
        logger.info('Passkey user is present and verified.')
    except Exception as exception:
        return _response(False, str(exception))

    try:
        credential_data = authenticator_data.credential_data
        UserPasskey.objects.create(
            user=user,
            name=_passkey_get_next_name(user),
            domain=domain,
            signature_counter=authenticator_data.counter,
            registration_flags=authenticator_data.flags,
            extensions=authenticator_data.extensions,
            aaguid=uuid.UUID(f'urn:uuid:{credential_data.aaguid}'),
            credential_id=credential_data.credential_id,
            public_key=fido2.cbor.encode(credential_data.public_key),
        )
    except IntegrityError:
        return _response(False,
                         _('Passkey with that identifier already exists.'))

    return _response(True, None)


@method_decorator(require_owner_or_admin, name='dispatch')
class PasskeyEdit(ContextMixin, UpdateView):
    """View to allow editing a passkey's name."""
    model = UserPasskey
    fields = ['name']
    pk_url_kwarg = 'passkey_id'
    title = _('Edit Passkey')

    def get_template_names(self):
        """Return the template name to use."""
        return ['users_passkey_edit.html']

    def get_success_url(self):
        """Return the URL to visit if form edit succeeds."""
        return reverse('users:passkeys', args=[self.kwargs['username']])


@method_decorator(require_owner_or_admin, name='dispatch')
@method_decorator(require_POST, name='dispatch')
class PasskeyDelete(DeleteView):
    """View to delete a passkey."""
    model = UserPasskey
    pk_url_kwarg = 'passkey_id'

    def get_success_url(self):
        """Return the URL to visit if form edit succeeds."""
        return reverse('users:passkeys', args=[self.kwargs['username']])


@json_exception
@require_POST
def passkey_login_begin(request):
    """Begin the process of logging-in with passwords."""
    # Domain
    domain = request.get_host().partition(':')[0]

    # Begin Authentication
    server = get_fido2_server(domain)
    request_options, state = server.authenticate_begin(
        credentials=None,
        user_verification=fido2.webauthn.UserVerificationRequirement.REQUIRED,
        challenge=None)

    logger.info('Passkey login begins')

    request.session['fido2_server_state'] = state
    return JsonResponse(dict(request_options))


@json_exception
@require_POST
def passkey_login_complete(request):
    """Complete the process of logging-in with passwords."""

    def _response(result: bool, error_string: str):
        """Return a JsonResponse object."""
        status = 200
        if not result:
            status = 400
            logger.error('Error completing passkey login: %s', error_string)

        return JsonResponse({
            'result': result,
            'error_string': error_string
        }, status=status)

    try:
        response = json.loads(request.body)
    except json.decoder.JSONDecodeError as exception:
        return _response(False, str(exception))

    # Domain
    domain = request.get_host().partition(':')[0]

    # State
    state = request.session.get('fido2_server_state')

    # Complete Authentication
    server = get_fido2_server(domain)
    try:
        authentication_repsonse = AuthenticationResponse.from_dict(response)

        if hasattr(authentication_repsonse, 'raw_id'):
            # Library python3-fido2 >= 2.0.0
            credential_id = authentication_repsonse.raw_id
        else:
            # Library python3-fido2 < 2.0.0
            credential_id = authentication_repsonse.id

        selected_passkeys = UserPasskey.objects.filter(
            credential_id=credential_id)

        if not len(selected_passkeys):
            return _response(False, _('Passkey used is not known.'))

        selected_passkey = selected_passkeys[0]

        credentials = [
            AttestedCredentialData.create(
                aaguid=selected_passkey.aaguid.bytes,
                credential_id=selected_passkey.credential_id,
                public_key=fido2.cbor.decode(selected_passkey.public_key))
        ]
        credential_data = server.authenticate_complete(
            state=state, credentials=credentials,
            response=authentication_repsonse)
    except Exception as exception:
        return _response(False, str(exception))

    try:
        passkeys = UserPasskey.objects.filter(
            credential_id=credential_data.credential_id,
            public_key=fido2.cbor.encode(credential_data.public_key))
    except Exception as exception:
        return _response(False, str(exception))

    if not len(passkeys):
        return _response(False, _('Passkey used is not known.'))

    assert len(passkeys) == 1  # credential_id is a unique field.

    passkey = passkeys[0]

    # Detect cloned passkeys using signature counter.
    # See: https://www.w3.org/TR/webauthn/#signature-counter
    authenticator_data = authentication_repsonse.response.authenticator_data
    signature_counter = authenticator_data.counter
    if (passkey.signature_counter and signature_counter
            and signature_counter <= passkey.signature_counter):
        # TODO: Notify user of a cloned passkey
        logger.warning(
            'Potentially cloned passkey detected. Passkey ID in DB - %s, '
            'credential ID - %s, signature counters - %s <= %s', passkey.id,
            fido2.utils.websafe_encode(passkey.credential_id),
            signature_counter, passkey.signature_counter)

    passkey.signature_counter = signature_counter
    passkey.save()  # Update the last used time

    # Needed by login(), stored in session, and used for permission checks.
    passkey.user.backend = 'django.contrib.auth.backends.ModelBackend'

    # Perform user login into Django
    auth_login(request, passkey.user)
    response = _response(True, None)
    if request.user.is_authenticated:
        translation.set_language(request, response,
                                 request.user.userprofile.language)

    return response



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to the Privacy app."""

import logging
import subprocess
from typing import Literal

from django.utils.translation import gettext_lazy as _
from django.utils.translation import gettext_noop

from plinth import app as app_module
from plinth import kvstore, menu
from plinth.config import DropinConfigs
from plinth.modules.backups.components import BackupRestore
from plinth.package import Packages

from . import manifest, privileged

logger = logging.getLogger(__name__)

_description = [_('Manage system-wide privacy settings.')]


class PrivacyApp(app_module.App):
    """FreedomBox app for Privacy."""

    app_id = 'privacy'

    _version = 4

    can_be_disabled = False

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               is_essential=True, name=_('Privacy'),
                               icon='fa-eye-slash', description=_description,
                               manual_page=None, tags=manifest.tags)
        self.add(info)

        menu_item = menu.Menu('menu-privacy', info.name, info.icon, info.tags,
                              'privacy:index', parent_url_name='system:data',
                              order=10)
        self.add(menu_item)

        packages = Packages('packages-privacy', ['popularity-contest', 'gpg'])
        self.add(packages)

        dropin_configs = DropinConfigs('dropin-configs-privacy', [
            '/etc/dpkg/origins/freedombox',
        ])
        self.add(dropin_configs)

        backup_restore = BackupRestore('backup-restore-privacy',
                                       **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup()
        if old_version == 0:
            privileged.set_configuration(enable_popcon=True)

        if old_version < 4:
            _show_privacy_notification()


def _show_privacy_notification():
    """Show a notification asking user to review privacy settings."""
    from plinth.notification import Notification
    message = gettext_noop(
        'Please update privacy settings to match your preferences.')
    data = {
        'app_name': 'translate:' + gettext_noop('Privacy'),
        'app_icon': 'fa-eye-slash'
    }
    title = gettext_noop('Review privacy setting')
    actions_ = [{
        'type': 'link',
        'class': 'primary',
        'text': gettext_noop('Go to {app_name}'),
        'url': 'privacy:index'
    }, {
        'type': 'dismiss'
    }]
    Notification.update_or_create(id='privacy-review', app_id='privacy',
                                  severity='info', title=title,
                                  message=message, actions=actions_, data=data,
                                  group='admin', dismissed=False)


def get_ip_lookup_url():
    """Return the URL to use to lookup external IP address."""
    return kvstore.get_default('ip_lookup_url',
                               'https://ddns.freedombox.org/ip/')


def set_ip_lookup_url(ip_lookup_url: str):
    """Set the URL to use to lookup external IP address."""
    kvstore.set('ip_lookup_url', ip_lookup_url)


def lookup_public_address(ip_type: Literal['ipv4', 'ipv6']) -> str:
    """Return the IP of this server/network by querying an external server."""
    lookup_url = get_ip_lookup_url()
    if not lookup_url:
        raise RuntimeError('Lookup URL not configured')

    ip_option = '-6' if ip_type == 'ipv6' else '-4'
    try:
        ip_address = subprocess.check_output([
            'wget', ip_option, '-o', '/dev/null', '-t', '3', '-T', '3', '-O',
            '-', lookup_url
        ])
        return ip_address.decode().strip().lower()
    except subprocess.CalledProcessError as exception:
        logger.warning('Unable to lookup external IP with URL %s: %s',
                       lookup_url, exception)
        message_map = {
            4: 'Network failure',
            5: 'SSL error',
            8: 'Server reponded with error'
        }
        raise RuntimeError(
            message_map.get(exception.returncode, 'Unknown error'))



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Application manifest for names.
"""

from django.utils.translation import gettext_lazy as _

backup: dict = {}

tags = [
    _('Domains'),
    _('Hostname'),
    _('DNS Resolution'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""URLs for the TiddlyWiki app."""

from django.urls import re_path

from .views import (CreateWikiView, TiddlyWikiAppView, RenameWikiView,
                    UploadWikiView, delete)

urlpatterns = [
    re_path(r'^apps/tiddlywiki/$', TiddlyWikiAppView.as_view(),
            name='index'),
    re_path(r'^apps/tiddlywiki/create/$', CreateWikiView.as_view(),
            name='create'),
    re_path(r'^apps/tiddlywiki/upload/$', UploadWikiView.as_view(),
            name='upload'),
    re_path(r'^apps/tiddlywiki/(?P<old_name>.+\.html)/rename/$',
            RenameWikiView.as_view(), name='rename'),
    re_path(r'^apps/tiddlywiki/(?P<name>.+\.html)/delete/$', delete,
            name='delete'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Set time zone with timedatectl."""

from plinth import action_utils
from plinth.actions import privileged


@privileged
def set_timezone(timezone: str):
    """Set time zone with timedatectl."""
    command = ['timedatectl', 'set-timezone', timezone]
    action_utils.run(command, check=True)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Functional, browser based tests for Feather Wiki app."""

import pathlib

import pytest

from plinth.tests import functional

pytestmark = [pytest.mark.apps, pytest.mark.featherwiki]

course_1 = 'Computer Organization and Architecture'
file_name_1 = 'Computer_Organization_and_Architecture.html'


class TestFeatherWikiApp(functional.BaseAppTests):
    app_name = 'featherwiki'
    has_service = False
    has_web = True

    def _create_wiki_file(self, session_browser):
        """Add a wiki to using the 'Create' functionality."""
        functional.nav_to_module(session_browser, 'featherwiki')

        wiki_link_1 = f'/featherwiki/{file_name_1}'
        if self._get_links_in_app_page(session_browser, wiki_link_1):
            return

        session_browser.links.find_by_href(
            '/freedombox/apps/featherwiki/create/').first.click()
        session_browser.find_by_id('id_featherwiki-name').fill(course_1)
        functional.submit(session_browser, form_class='form-featherwiki')

    def _get_links_in_app_page(self, session_browser, link):
        """Return the links matching a href in the app page."""
        functional.nav_to_module(session_browser, 'featherwiki')
        return session_browser.links.find_by_href(link)

    def _get_links_in_apache_listing(self, session_browser, link):
        """Return the links matching a href in the index page."""
        default_url = functional.config['DEFAULT']['url']
        session_browser.visit(f'{default_url}/featherwiki')
        return session_browser.links.find_by_href(link)

    def _assert_wiki_present(self, session_browser, file_name, present=True):
        """Assert that a wiki is present."""
        wiki_link = f'/featherwiki/{file_name}'
        assert bool(self._get_links_in_app_page(session_browser,
                                                wiki_link)) == present
        assert bool(
            self._get_links_in_apache_listing(session_browser,
                                              file_name)) == present

    def _assert_wiki_works(self, session_browser, file_name):
        """Assert that wiki loads and run as expected."""
        wiki_link = f'/featherwiki/{file_name}'
        default_url = functional.config['DEFAULT']['url']
        session_browser.visit(f'{default_url}{wiki_link}')
        links = session_browser.links.find_by_href('https://feather.wiki')
        assert len(links) == 1

    def test_wiki_file_access(self, session_browser):
        """Test creating a new wiki file."""
        self._create_wiki_file(session_browser)

        self._assert_wiki_present(session_browser, file_name_1)
        self._assert_wiki_works(session_browser, file_name_1)

    def test_rename_wiki_file(self, session_browser):
        """Test changing the name of a wiki file."""
        self._create_wiki_file(session_browser)

        new_course = 'A Midsummer Night\'s Dream'
        new_file_name = 'A_Midsummer_Nights_Dream.html'
        self._get_links_in_app_page(
            session_browser, '/freedombox/apps/featherwiki/' + file_name_1 +
            '/rename/').first.click()
        session_browser.find_by_id('id_featherwiki-new_name').fill(new_course)
        functional.submit(session_browser, form_class='form-featherwiki')

        self._assert_wiki_present(session_browser, new_file_name)
        self._assert_wiki_works(session_browser, new_file_name)

    def test_upload_wiki_file(self, session_browser):
        """Test uploading an existing wiki file."""
        _test_data_dir = pathlib.Path(__file__).parent / 'data'
        test_wiki_file = str(_test_data_dir / 'dummy_wiki.html')

        session_browser.links.find_by_href(
            '/freedombox/apps/featherwiki/upload/').first.click()
        session_browser.attach_file('featherwiki-file', test_wiki_file)
        functional.submit(session_browser, form_class='form-featherwiki')

        self._assert_wiki_present(session_browser, 'dummy_wiki.html')

    def test_delete_wiki_file(self, session_browser):
        """Test deleting an existing wiki file"""
        self._create_wiki_file(session_browser)

        self._get_links_in_app_page(
            session_browser, '/freedombox/apps/featherwiki/' + file_name_1 +
            '/delete/').first.click()
        functional.submit(session_browser, form_class='form-delete')

        self._assert_wiki_present(session_browser, file_name_1, present=False)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Tests for Miniflux views."""

from unittest.mock import patch

import pytest
from django import urls
from django.contrib.messages.storage.fallback import FallbackStorage

from plinth import module_loader
from plinth.modules.miniflux import views

# For all tests, use plinth.urls instead of urls configured for testing
pytestmark = pytest.mark.urls('plinth.urls')


@pytest.fixture(autouse=True, scope='module')
def fixture_miniflux_urls():
    """Make sure Miniflux app's URLs are part of plinth.urls."""
    with patch('plinth.module_loader._modules_to_load', new=[]) as modules, \
            patch('plinth.urls.urlpatterns', new=[]):
        modules.append('plinth.modules.miniflux')
        module_loader.include_urls()
        yield


def make_request(request, view, **kwargs):
    """Make request with a message storage."""
    setattr(request, 'session', 'session')
    messages = FallbackStorage(request)
    setattr(request, '_messages', messages)
    response = view(request, **kwargs)

    return response, messages


##########################
# Create Admin User view #
##########################


def test_create_admin_user_view(rf):
    """Test that the create admin user view loads successfully."""
    request = rf.get(urls.reverse('miniflux:create-admin-user'))
    view = views.CreateAdminUserView.as_view()
    response, _ = make_request(request, view)

    assert response.status_code == 200


@patch('plinth.modules.miniflux.privileged.create_admin_user')
def test_create_admin_user_form_valid(create_admin_user, rf):
    """Test that the create admin user form is valid and redirects."""
    form_data = {
        'miniflux-username': 'admin',
        'miniflux-password': 'strongpassword',
        'miniflux-password_confirmation': 'strongpassword'
    }
    request = rf.post(urls.reverse('miniflux:create-admin-user'),
                      data=form_data)
    view = views.CreateAdminUserView.as_view()
    response, messages = make_request(request, view)

    assert response.status_code == 302
    assert list(messages)[0].message == 'Created admin user: admin'


def test_passwords_do_not_match(rf):
    """Test that the form shows an error when passwords do not match."""
    form_data = {
        'miniflux-username': 'admin',
        'miniflux-password': 'strongpassword',
        'miniflux-password_confirmation': 'weakpassword'
    }
    request = rf.post(urls.reverse('miniflux:create-admin-user'),
                      data=form_data)
    view = views.CreateAdminUserView.as_view()
    response, messages = make_request(request, view)

    assert response.status_code == 200
    assert response.context_data['form'].errors['password_confirmation'][
        0] == 'Passwords do not match.'


def test_password_too_short(rf):
    """Test that the form shows an error when the password is too short."""
    form_data = {
        'miniflux-username': 'demo',
        'miniflux-password': 'demo',
        'miniflux-password_confirmation': 'demo'
    }
    request = rf.post(urls.reverse('miniflux:create-admin-user'),
                      data=form_data)
    view = views.CreateAdminUserView.as_view()
    response, messages = make_request(request, view)

    assert response.status_code == 200
    assert response.context_data['form'].errors['password'][
        0] == 'Ensure this value has at least 6 characters (it has 4).'


@patch('plinth.modules.miniflux.privileged.create_admin_user')
def test_recreate_existing_user(create_admin_user, rf):
    """Test that trying to recreate an existing user fails."""
    create_admin_user.side_effect = Exception(
        'Skipping admin user creation because it already exists')

    form_data = {
        'miniflux-username': 'admin',
        'miniflux-password': 'strongpassword',
        'miniflux-password_confirmation': 'strongpassword'
    }
    request = rf.post(urls.reverse('miniflux:create-admin-user'),
                      data=form_data)
    view = views.CreateAdminUserView.as_view()
    response, messages = make_request(request, view)

    error_msg = ('An error occurred while creating the user: Skipping admin '
                 'user creation because it already exists.')
    assert response.status_code == 302
    assert list(messages)[0].message == error_msg


############################
# Reset User Password view #
############################


@patch('plinth.modules.miniflux.privileged.reset_user_password')
def test_reset_user_password_form_valid(reset_user_password, rf):
    """Test that the reset user password form is valid and redirects."""
    reset_user_password.return_value = 'Password changed!'

    form_data = {
        'miniflux-username': 'admin',
        'miniflux-password': 'strongpassword',
        'miniflux-password_confirmation': 'strongpassword'
    }
    request = rf.post(urls.reverse('miniflux:reset-user-password'),
                      data=form_data)
    view = views.ResetUserPasswordView.as_view()
    response, messages = make_request(request, view)

    assert response.status_code == 302
    assert list(messages)[0].message == 'Password reset for user: admin'


@patch('plinth.modules.miniflux.privileged.reset_user_password')
def test_reset_user_password_for_invalid_user(reset_user_password, rf):
    """Test that the resetting user password for an invalid user fails."""
    reset_user_password.side_effect = Exception('user not found')

    form_data = {
        'miniflux-username': 'admin',
        'miniflux-password': 'strongpassword',
        'miniflux-password_confirmation': 'strongpassword'
    }
    request = rf.post(urls.reverse('miniflux:reset-user-password'),
                      data=form_data)
    view = views.ResetUserPasswordView.as_view()
    response, messages = make_request(request, view)

    assert response.status_code == 302
    assert list(
        messages
    )[0].message == 'An error occurred during password reset: user not found.'



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Application manifest for firewall.
"""

from django.utils.translation import gettext_lazy as _

backup: dict = {}

tags = [_('Ports'), _('Blocking'), _('Status'), _('Automatic')]



# SPDX-License-Identifier: AGPL-3.0-or-later

from typing import ClassVar

from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from plinth import app


class Menu(app.FollowerComponent):
    """Component to manage a single menu item."""

    _all_menus: ClassVar[set['Menu']] = set()

    def __init__(self, component_id: str, name: str | None = None,
                 icon: str | None = None, tags: list[str] | None = None,
                 url_name: str | None = None, url_args: list | None = None,
                 url_kwargs: dict | None = None,
                 parent_url_name: str | None = None, order: int = 50,
                 advanced: bool = False):
        """Initialize a new menu item with basic properties.

        name is the label of the menu item.

        icon is the icon to be displayed for the menu item. Icon can be the
        name of an SVG icon from the static/theme/icons directory. In this
        case, the icon name starts with the string 'fa-' (a tribute to the
        fork-awesome project). Alternatively, the icon can also be a file under
        the directory plinth/modules/<app>/static/icons/, provided without an
        extension. SVG icons are preferred. Currently, both PNG and SVG icons
        with the same name are used. For example, if the value of icon is
        'myicon' and app_id in App class is 'myapp', then two icons files
        plinth/modules/myapp/static/icons/myicon.svg and
        plinth/modules/myapp/static/icons/myicon.png are used in the interface.

        tags is a list of tags that describe the app. Tags help users to find
        similar apps or alternatives and discover use cases.

        url_name is the name of url location that will be activated when the
        menu item is selected. This is not optional. url_args and url_kwargs
        are sent to reverse() when resolving url from url_name.

        parent_url_name optionally specifies the menu item under which this
        menu item should become a child.

        order is the numerical rank of this item within the menu. Lower order
        items appear closest to the top/left of the menu. By convention, we use
        the spectrum between 0 and 100 to rank orders, but feel free to
        disregard that. If you need more granularity, don't bother renumbering
        things. Feel free to use fractional orders.

        advanced decides whether to show the menu item only in advanced mode.

        """
        super().__init__(component_id)
        if not url_name:
            raise ValueError('Valid url_name is expected')

        url = reverse_lazy(url_name, args=url_args, kwargs=url_kwargs)

        self.name = name
        self.icon = icon
        self.tags = tags
        self.url = url
        self.order = order
        self.advanced = advanced

        self.url_name = url_name
        self.url_args = url_args
        self.url_kwargs = url_kwargs
        self.parent_url_name = parent_url_name

        # Add self to global list of menu items.
        self._all_menus.add(self)

    @property
    def items(self):
        """Return the list of children for this menu item."""
        return [
            item for item in self._all_menus
            if item.parent_url_name == self.url_name
        ]

    def sorted_items(self):
        """Return menu items in sorted order according to current locale."""
        return sorted(self.items, key=lambda x: (x.order, x.name.lower()))

    def active_item(self, request):
        """Return the first active item (e.g. submenu) that is found."""
        for item in self.items:
            if request.path.startswith(str(item.url)):
                return item

        return None

    @staticmethod
    def get_with_url_name(url_name: str) -> 'Menu':
        """Return a menu item with given URL name.

        Raise LookupError of the request item is not found.
        """
        for item in Menu._all_menus:
            if item.url_name == url_name:
                return item

        raise LookupError


main_menu = None


def init():
    """Create main menu and other essential menus."""
    global main_menu
    main_menu = Menu('menu-index', name=_('Home'), url_name='index')
    Menu('menu-apps', name=_('Apps'), icon='fa-download', url_name='apps',
         parent_url_name='index')
    Menu('menu-system', name=_('System'), icon='fa-cog', url_name='system',
         parent_url_name='index')

    Menu('menu-system-visibility', name=_('Visibility'), icon='fa-cog',
         url_name='system:visibility', parent_url_name='system', order=10)
    Menu('menu-system-data', name=_('Data'), icon='fa-cog',
         url_name='system:data', parent_url_name='system', order=20)
    Menu('menu-system-system', name=_('System'), icon='fa-cog',
         url_name='system:system', parent_url_name='system', order=30)
    Menu('menu-system-security', name=_('Security'), icon='fa-cog',
         url_name='system:security', parent_url_name='system', order=40)
    Menu('menu-system-administration', name=_('Administration'), icon='fa-cog',
         url_name='system:administration', parent_url_name='system', order=50)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app for bepasty."""

from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import frontpage, menu
from plinth.config import DropinConfigs
from plinth.daemon import Daemon
from plinth.modules.apache.components import Webserver
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import Firewall
from plinth.package import Packages
from plinth.privileged import service as service_privileged

from . import manifest, privileged

_description = [
    _('bepasty is a web application that allows large files to be uploaded '
      'and shared. Text and code snippets can also be pasted and shared. '
      'Text, image, audio, video and PDF documents can be previewed in the '
      'browser. Shared files can be set to expire after a time period.'),
    _('bepasty does not use usernames for login. It only uses passwords. For '
      'each password, a set of permissions can be selected. Once you have '
      'created a password, you can share it with the users who should have the'
      ' associated permissions.'),
    _('You can also create multiple passwords with the same set of privileges,'
      ' and distribute them to different people or groups. This will allow '
      'you to later revoke access for a single person or group, by removing '
      'their password from the list.'),
]

PERMISSIONS = {
    'read': _('Read a file, if a web link to the file is available'),
    'create': _('Create or upload files'),
    'list': _('List all files and their web links'),
    'delete': _('Delete files'),
    'admin': _('Administer files: lock/unlock files'),
}

DEFAULT_PERMISSIONS = {
    '': _('None, password is always required'),
    'read': _('Read a file, if a web link to the file is available'),
    'read list': _('List and read all files'),
}


class BepastyApp(app_module.App):
    """FreedomBox app for bepasty."""

    app_id = 'bepasty'

    _version = 4

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(self.app_id, self._version, name=_('bepasty'),
                               icon_filename='bepasty',
                               description=_description, manual_page='bepasty',
                               clients=manifest.clients, tags=manifest.tags)
        self.add(info)

        menu_item = menu.Menu('menu-bepasty', info.name, info.icon_filename,
                              info.tags, 'bepasty:index',
                              parent_url_name='apps')
        self.add(menu_item)

        shortcut = frontpage.Shortcut('shortcut-bepasty', info.name,
                                      info.icon_filename, '/bepasty',
                                      clients=manifest.clients, tags=info.tags)
        self.add(shortcut)

        packages = Packages('packages-bepasty', ['bepasty'])
        self.add(packages)

        dropin_configs = DropinConfigs('dropin-configs-bepasty', [
            '/etc/apache2/conf-available/bepasty-freedombox.conf',
            '/etc/uwsgi/apps-available/bepasty-freedombox.ini'
        ])
        self.add(dropin_configs)

        firewall = Firewall('firewall-bepasty', info.name,
                            ports=['http', 'https'], is_external=True)
        self.add(firewall)

        daemon = Daemon('daemon-bepasty',
                        'uwsgi-app@bepasty-freedombox.socket')
        self.add(daemon)

        webserver = Webserver('webserver-bepasty', 'bepasty-freedombox',
                              urls=['https://{host}/bepasty/'])
        self.add(webserver)

        backup_restore = BackupRestore('backup-restore-bepasty',
                                       **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup('freedombox.local')
        if not old_version:
            self.enable()

        if old_version == 1 and not privileged.get_configuration().get(
                'DEFAULT_PERMISSIONS'):
            # Upgrade to a better default only if user hasn't changed the
            # value.
            privileged.set_default(['read'])

        if old_version and old_version <= 3:
            webserver = self.get_component('webserver-bepasty')
            daemon = self.get_component('daemon-bepasty')
            if webserver.is_enabled():
                daemon.enable()

            # Vanquish the old uwsgi init.d script.
            service_privileged.disable('uwsgi')
            service_privileged.mask('uwsgi')

    def uninstall(self):
        """De-configure and uninstall the app."""
        super().uninstall()
        privileged.uninstall()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to manage users."""

import grp
import subprocess

import augeas
from django.utils.text import format_lazy
from django.utils.translation import gettext_lazy as _
from django.utils.translation import gettext_noop

from plinth import action_utils
from plinth import app as app_module
from plinth import cfg, menu
from plinth.config import DropinConfigs
from plinth.daemon import Daemon
from plinth.diagnostic_check import (DiagnosticCheck,
                                     DiagnosticCheckParameters, Result)
from plinth.package import Packages
from plinth.privileged import service as service_privileged

from . import manifest, privileged
from .components import UsersAndGroups

first_boot_steps = [
    {
        'id': 'users_firstboot',
        'url': 'users:firstboot',
        'order': 1
    },
]

_description = [
    _('Create and manage user accounts. These accounts serve as centralized '
      'authentication mechanism for most apps. Some apps further require a '
      'user account to be part of a group to authorize the user to access the '
      'app.'),
    format_lazy(
        _('Any user may login to {box_name} web interface to see a list of '
          'apps relevant to them in the home page. However, only users of '
          'the <em>admin</em> group may alter apps or system settings.'),
        box_name=_(cfg.box_name))
]


class UsersApp(app_module.App):
    """FreedomBox app for users and groups management."""

    app_id = 'users'

    _version = 8

    can_be_disabled = False

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               is_essential=True, name=_('Users and Groups'),
                               icon='fa-users', description=_description,
                               manual_page='Users', tags=manifest.tags)
        self.add(info)

        menu_item = menu.Menu('menu-users', info.name, info.icon, info.tags,
                              'users:index', parent_url_name='system:system',
                              order=10)
        self.add(menu_item)

        packages = Packages('packages-users', [
            'ldapscripts', 'ldap-utils', 'libnss-ldapd', 'libpam-ldapd',
            'nslcd', 'samba-common-bin', 'slapd', 'tdb-tools'
        ])
        self.add(packages)

        dropin_configs = DropinConfigs(
            'dropin-configs-users',
            [
                '/etc/apache2/includes/freedombox-auth-ldap.conf',
                # Empty file kept for easier upgrade
                '/etc/apache2/includes/freedombox-single-sign-on.conf',
            ])
        self.add(dropin_configs)

        daemon = Daemon('daemon-users', 'slapd', listen_ports=[(389, 'tcp4'),
                                                               (389, 'tcp6')])
        self.add(daemon)

        # Add the admin group
        groups = {'admin': _('Access to all services and system settings')}
        users_and_groups = UsersAndGroups('users-and-groups-admin',
                                          groups=groups)
        self.add(users_and_groups)

    def diagnose(self) -> list[DiagnosticCheck]:
        """Run diagnostics and return the results."""
        results = super().diagnose()

        results.append(_diagnose_ldap_entry('dc=thisbox'))
        results.append(_diagnose_ldap_entry('ou=users'))
        results.append(_diagnose_ldap_entry('ou=groups'))
        results.append(_diagnose_ldap_entry('ou=policies'))
        results.append(_diagnose_ldap_entry('cn=DefaultPPolicy'))

        config = privileged.get_nslcd_config()
        results.append(_diagnose_nslcd_config(config, 'uri', 'ldapi:///'))
        results.append(_diagnose_nslcd_config(config, 'base', 'dc=thisbox'))
        results.append(_diagnose_nslcd_config(config, 'sasl_mech', 'EXTERNAL'))
        results.extend(_diagnose_nsswitch_config())

        return results

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        if not old_version:
            privileged.first_setup()

        if old_version and old_version < 7:
            # Setup password policy and lock LDAP passwords for inactive users.
            inactivated_users = _get_inactivated_users()
            if inactivated_users:
                privileged.setup_and_sync_user_states(inactivated_users)

        privileged.setup()
        privileged.create_group('freedombox-share')


def _diagnose_ldap_entry(search_item: str) -> DiagnosticCheck:
    """Diagnose that an LDAP entry exists."""
    check_id = f'users-ldap-entry-{search_item}'
    result = Result.FAILED

    try:
        output = action_utils.run(
            ['ldapsearch', '-LLL', '-x', '-b', 'dc=thisbox', search_item],
            check=True).stdout
        if search_item in output.decode():
            result = Result.PASSED
    except subprocess.CalledProcessError:
        pass

    description = gettext_noop('Check LDAP entry "{search_item}"')
    parameters: DiagnosticCheckParameters = {'search_item': search_item}

    return DiagnosticCheck(check_id, description, result, parameters)


def _diagnose_nslcd_config(config: dict[str, str], key: str,
                           value: str) -> DiagnosticCheck:
    """Diagnose that nslcd has a configuration."""
    check_id = f'users-nslcd-config-{key}'
    try:
        result = Result.PASSED if config[key] == value else Result.FAILED
    except KeyError:
        result = Result.FAILED

    description = gettext_noop('Check nslcd config "{key} {value}"')
    parameters: DiagnosticCheckParameters = {'key': key, 'value': value}

    return DiagnosticCheck(check_id, description, result, parameters)


def _diagnose_nsswitch_config() -> list[DiagnosticCheck]:
    """Diagnose that Name Service Switch is configured to use LDAP."""
    nsswitch_conf = '/etc/nsswitch.conf'
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.transform('Nsswitch', nsswitch_conf)
    aug.set('/augeas/context', '/files' + nsswitch_conf)
    aug.load()

    results = []
    for database in ['passwd', 'group', 'shadow']:
        check_id = f'users-nsswitch-config-{database}'
        result = Result.FAILED
        for match in aug.match('database'):
            if aug.get(match) != database:
                continue

            for service_match in aug.match(match + '/service'):
                if 'ldap' == aug.get(service_match):
                    result = Result.PASSED
                    break

            break

        description = gettext_noop('Check nsswitch config "{database}"')
        parameters: DiagnosticCheckParameters = {'database': database}

        results.append(
            DiagnosticCheck(check_id, description, result, parameters))

    return results


def get_last_admin_user():
    """If there is only one admin user return its name else return None."""
    admin_users = privileged.get_group_users('admin')
    if len(admin_users) == 1 and admin_users[0]:
        return admin_users[0]

    return None


def _get_inactivated_users() -> list[str]:
    """Get list of inactivated usernames"""
    from django.contrib.auth.models import User
    users = User.objects.filter(is_active=False)

    return [user.username for user in users]


def add_user_to_share_group(username, service=None):
    """Add user to the freedombox-share group."""
    try:
        group_members = grp.getgrnam('freedombox-share').gr_mem
    except KeyError:
        group_members = []
    if username not in group_members:
        privileged.add_user_to_group(username, 'freedombox-share')
        if service:
            service_privileged.try_restart(service)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for syncthing app.
"""

import time

import pytest

from plinth.tests import functional

pytestmark = [pytest.mark.apps, pytest.mark.syncthing]


class TestSyncthingApp(functional.BaseAppTests):
    app_name = 'syncthing'
    has_service = True
    has_web = True

    def test_notifications(self, session_browser):
        """Test that authentication and usage reporting notifications are not
        shown."""
        functional.app_enable(session_browser, self.app_name)
        functional.access_url(session_browser, self.app_name)
        _assert_usage_report_notification_not_shown(session_browser)
        _assert_authentication_notification_not_shown(session_browser)

    def test_add_remove_folder(self, session_browser):
        """Test adding and removing a folder."""
        functional.app_enable(session_browser, self.app_name)
        if _folder_is_present(session_browser, 'Test'):
            _remove_folder(session_browser, 'Test')

        _add_folder(session_browser, 'Test', '/tmp')
        assert _folder_is_present(session_browser, 'Test')

        _remove_folder(session_browser, 'Test')
        assert not _folder_is_present(session_browser, 'Test')

    @pytest.mark.backups
    def test_backup_restore(self, session_browser):
        """Test backup and restore of app data."""
        functional.app_enable(session_browser, self.app_name)
        if _folder_is_present(session_browser, 'Test'):
            _remove_folder(session_browser, 'Test')

        _add_folder(session_browser, 'Test', '/tmp')
        functional.backup_create(session_browser, self.app_name,
                                 'test_syncthing')

        _remove_folder(session_browser, 'Test')
        time.sleep(1)  # Helps with browsing away in next step
        functional.backup_restore(session_browser, self.app_name,
                                  'test_syncthing')

        assert _folder_is_present(session_browser, 'Test')

    def test_user_group_access(self, session_browser):
        """Test that only users in syncthing-access group can access syncthing
        site."""
        functional.app_enable(session_browser, self.app_name)
        if not functional.user_exists(session_browser, 'syncthinguser'):
            functional.create_user(session_browser, 'syncthinguser',
                                   groups=['syncthing-access'])
        if not functional.user_exists(session_browser, 'nogroupuser'):
            functional.create_user(session_browser, 'nogroupuser')

        functional.login_with_account(session_browser, functional.base_url,
                                      'syncthinguser')
        assert functional.is_available(session_browser, self.app_name)

        functional.login_with_account(session_browser, functional.base_url,
                                      'nogroupuser')
        assert not functional.is_available(session_browser, self.app_name)

        functional.login(session_browser)


def _assert_usage_report_notification_not_shown(session_browser):
    _load_main_interface(session_browser)
    assert session_browser.find_by_id('ur').visible is False


def _assert_authentication_notification_not_shown(session_browser):
    _load_main_interface(session_browser)
    assert bool(session_browser.find_by_css(
        '#authenticationUserAndPassword *')) is False


def _load_main_interface(browser):
    """Close the dialog boxes that many popup after visiting the URL."""
    functional.access_url(browser, 'syncthing')

    def service_is_available():
        if browser.is_element_present_by_xpath(
                '//h1[text()="Service Unavailable"]'):
            functional.access_url(browser, 'syncthing')
            return False

        return True

    # After a backup restore, service may not be available immediately
    functional.eventually(service_is_available)

    # Wait for javascript loading process to complete
    functional.eventually(lambda: browser.evaluate_script(
        'angular.element("[ng-controller=SyncthingController]").scope()'
        '.thisDevice().name'))

    # Give browser additional time to setup site
    time.sleep(1)


def _folder_is_present(browser, folder_name):
    """Return whether a folder is present in Syncthing."""
    _load_main_interface(browser)
    folder_names = browser.find_by_css('#folders .panel-title-text span')
    folder_names = [folder_name.text for folder_name in folder_names]
    return folder_name in folder_names


def _add_folder(browser, folder_name, folder_path):
    """Add a new folder to Synthing."""
    _load_main_interface(browser)
    add_folder_xpath = '//button[contains(@ng-click, "addFolder")]'
    browser.find_by_xpath(add_folder_xpath).click()

    folder_dialog = browser.find_by_id('editFolder').first
    functional.eventually(lambda: folder_dialog.visible)
    browser.find_by_id('folderLabel').fill(folder_name)
    browser.find_by_id('folderPath').fill(folder_path)
    save_folder_xpath = './/button[contains(@ng-click, "saveFolder")]'
    folder_dialog.find_by_xpath(save_folder_xpath).first.click()
    functional.eventually(lambda: not folder_dialog.visible)


def _remove_folder(browser, folder_name):
    """Remove a folder from Synthing."""
    _load_main_interface(browser)

    # Find folder
    folder = None
    for current_folder in browser.find_by_css('#folders > .panel'):
        name = current_folder.find_by_css('.panel-title-text span').first.text
        if name == folder_name:
            folder = current_folder
            break

    # Edit folder button
    folder.find_by_css('button.panel-heading').first.click()
    functional.eventually(lambda: folder.find_by_css('div.collapse.in'))
    edit_folder_xpath = './/button[contains(@ng-click, "editFolder")]'
    edit_folder_button = folder.find_by_xpath(edit_folder_xpath).first
    edit_folder_button.scroll_to()
    edit_folder_button.click()

    # Edit folder dialog
    folder_dialog = browser.find_by_id('editFolder').first
    functional.eventually(lambda: folder_dialog.visible)
    remove_button_xpath = './/button[contains(@data-target, "remove-folder")]'
    folder_dialog.find_by_xpath(remove_button_xpath).first.click()

    # Remove confirmation dialog
    remove_folder_dialog = browser.find_by_id('remove-folder-confirmation')
    functional.eventually(lambda: remove_folder_dialog.visible)
    remove_button_xpath = './/button[contains(@ng-click, "deleteFolder")]'
    remove_folder_dialog.find_by_xpath(remove_button_xpath).first.click()

    functional.eventually(lambda: not folder_dialog.visible)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Tests for CherryPy web server setup and its components.
"""

import pathlib
from unittest.mock import Mock, call, patch

import pytest

from plinth.web_server import StaticFiles, resolve_static_path


@pytest.fixture(autouse=True)
def fixture_cleanup_static_files():
    """Ensure that global list of static files is clean."""
    StaticFiles._all_instances = {}


def test_static_files_init():
    """Test that static files component is being initialized correctly."""
    component = StaticFiles('test-component')
    assert component.component_id == 'test-component'
    assert component.directory_map is None

    directory_map = {'/a': '/b'}
    component = StaticFiles('test-component', directory_map)
    assert component.directory_map == directory_map


def test_static_files_list():
    """Test that static files components can be listed properly."""
    component1 = StaticFiles('test-component1')
    component2 = StaticFiles('test-component2')

    assert set(StaticFiles.list()) == {component1, component2}


@patch('cherrypy.tree.mount')
def test_static_files_mount(mount, load_cfg):
    """Test that mounting on CherryPy works as expected."""
    directory_map = {'/a': '/b', '/c': '/d'}
    component = StaticFiles('test-component', directory_map)
    component.mount()

    calls = [
        call(
            None, '/freedombox/a', {
                '/': {
                    'tools.staticdir.root': '/b',
                    'tools.staticdir.on': True,
                    'tools.staticdir.dir': '.'
                }
            }),
        call(
            None, '/freedombox/c', {
                '/': {
                    'tools.staticdir.root': '/d',
                    'tools.staticdir.on': True,
                    'tools.staticdir.dir': '.'
                }
            })
    ]
    mount.assert_has_calls(calls)


@patch('sys.modules')
@patch('plinth.app.App.list')
def test_resolve_static_path(app_list, sys_modules, tmp_path):
    """Test that resolving a static path works as expected."""
    app_list.return_value = []
    expected_path = (pathlib.Path(__file__).parent.parent.parent /
                     'static/theme/icons/test.svg')
    assert resolve_static_path('theme/icons/test.svg') == expected_path

    app = Mock()
    app.app_id = 'test-app'
    app.__module__ = 'test-module'
    app_list.return_value = [app]
    expected_path = (pathlib.Path(__file__).parent.parent.parent /
                     'static/theme/icons/test.svg')

    sys_modules.__getitem__.side_effect = {}.__getitem__
    assert resolve_static_path('theme/icons/test.svg') == expected_path
    with pytest.raises(ValueError, match='Module for app not loaded'):
        resolve_static_path('test-app/test.svg')

    module = Mock()
    sys_modules.__getitem__.side_effect = {'test-module': module}.__getitem__
    with pytest.raises(ValueError,
                       match='Module file for app could not be found'):
        resolve_static_path('test-app/test.svg')

    module.__file__ = tmp_path / 'test-module.py'
    with pytest.raises(ValueError,
                       match='No static directory available for app test-app'):
        resolve_static_path('test-app/test.svg')

    (tmp_path / 'static').mkdir()
    assert resolve_static_path(
        'test-app/test.svg') == tmp_path / 'static/test.svg'



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the Network module
"""

from django.urls import re_path

from . import views

urlpatterns = [
    re_path(r'^sys/networks/$', views.NetworksAppView.as_view(), name='index'),
    re_path(r'^sys/networks/(?P<uuid>[\w.@+-]+)/show/$', views.show,
            name='show'),
    re_path(r'^sys/networks/(?P<uuid>[\w.@+-]+)/edit/$', views.edit,
            name='edit'),
    re_path(r'^sys/networks/(?P<uuid>[\w.@+-]+)/activate/$', views.activate,
            name='activate'),
    re_path(r'^sys/networks/(?P<uuid>[\w.@+-]+)/deactivate/$',
            views.deactivate, name='deactivate'),
    re_path(r'^sys/networks/scan/$', views.scan, name='scan'),
    re_path(r'^sys/networks/add/$', views.add, name='add'),
    re_path(r'^sys/networks/add/generic/$', views.add_generic,
            name='add_generic'),
    re_path(r'^sys/networks/add/ethernet/$', views.add_ethernet,
            name='add_ethernet'),
    re_path(r'^sys/networks/add/pppoe/$', views.add_pppoe, name='add_pppoe'),
    re_path(
        r'^sys/networks/add/wifi/(?:(?P<ssid>[^/]+)/'
        r'(?P<interface_name>[^/]+)/)?$', views.add_wifi, name='add_wifi'),
    re_path(r'^sys/networks/(?P<uuid>[\w.@+-]+)/delete/$', views.delete,
            name='delete'),
    re_path(r'^sys/networks/router-configuration/$',
            views.RouterConfigurationView.as_view(),
            name='router-configuration'),
    re_path(r'^sys/networks/internet-connection-type/$',
            views.InternetConnectionTypeView.as_view(),
            name='internet-connection-type'),
    re_path(r'^sys/networks/network-topology/$',
            views.NetworkTopologyView.as_view(), name='network-topology'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to configure Mumble server."""

import pathlib
from typing import Iterator

from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _
from django.utils.translation import gettext_noop

from plinth import app as app_module
from plinth import frontpage, menu
from plinth.daemon import Daemon
from plinth.diagnostic_check import DiagnosticCheck, Result
from plinth.modules import names
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import Firewall
from plinth.modules.letsencrypt.components import LetsEncrypt
from plinth.modules.users.components import UsersAndGroups
from plinth.package import Packages, install
from plinth.utils import Version

from . import manifest, privileged

_description = [
    _('Mumble is an open source, low-latency, encrypted, high quality '
      'voice chat software.'),
    _('You can connect to your Mumble server on the regular Mumble port '
      '64738. <a href="http://mumble.info">Clients</a> to connect to Mumble '
      'from your desktop and mobile devices are available.')
]


class MumbleApp(app_module.App):
    """FreedomBox app for Mumble."""

    app_id = 'mumble'

    _version = 2

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(
            app_id=self.app_id, version=self._version, name=_('Mumble'),
            icon_filename='mumble', description=_description,
            manual_page='Mumble', clients=manifest.clients, tags=manifest.tags,
            donation_url='https://wiki.mumble.info/wiki/Donate')
        self.add(info)

        menu_item = menu.Menu('menu-mumble', info.name, info.icon_filename,
                              info.tags, 'mumble:index',
                              parent_url_name='apps')
        self.add(menu_item)

        shortcut = frontpage.Shortcut(
            'shortcut-mumble', info.name, icon=info.icon_filename,
            description=info.description, manual_page=info.manual_page,
            configure_url=reverse_lazy('mumble:index'), clients=info.clients,
            tags=info.tags)
        self.add(shortcut)

        packages = Packages('packages-mumble', ['mumble-server'])
        self.add(packages)

        firewall = Firewall('firewall-mumble', info.name,
                            ports=['mumble-plinth'], is_external=True)
        self.add(firewall)

        letsencrypt = LetsEncrypt(
            'letsencrypt-mumble', domains=get_domains,
            daemons=['mumble-server'], should_copy_certificates=True,
            private_key_path='/var/lib/mumble-server/privkey.pem',
            certificate_path='/var/lib/mumble-server/fullchain.pem',
            user_owner='mumble-server', group_owner='mumble-server',
            managing_app='mumble')
        self.add(letsencrypt)

        daemon = Daemon(
            'daemon-mumble', 'mumble-server', listen_ports=[(64738, 'tcp4'),
                                                            (64738, 'tcp6'),
                                                            (64738, 'udp4'),
                                                            (64738, 'udp6')])
        self.add(daemon)

        users_and_groups = UsersAndGroups('users-and-groups-mumble',
                                          reserved_usernames=['mumble-server'])
        self.add(users_and_groups)

        backup_restore = BackupRestore('backup-restore-mumble',
                                       **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup()
        if not old_version:
            self.enable()

        self.get_component('letsencrypt-mumble').setup_certificates()

    def force_upgrade(self, packages):
        """Force upgrade mumble-server to resolve conffile prompts."""
        if 'mumble-server' not in packages:
            return False

        # Allow upgrades from 1.3.4-1 (bullseye) to 1.3.4-4 (bookworm) but not
        # to 1.5.735 (trixie) and beyond.
        package = packages['mumble-server']
        if Version(package['new_version']) > Version('1.4~'):
            return False

        install(['mumble-server'], force_configuration='new')
        privileged.setup()
        return True

    def diagnose(self) -> list[DiagnosticCheck]:
        """Run diagnostics and return the results."""
        results = super().diagnose()
        results.append(_diagnose_config())
        return results


def get_available_domains() -> Iterator[str]:
    """Return an iterator with all domains able to have a certificate."""
    return (domain.name for domain in names.components.DomainName.list()
            if domain.domain_type.can_have_certificate)


def get_domain():
    """Read TLS domain from config file select first available if none."""
    domain = privileged.get_domain()
    if not domain:
        domain = next(get_available_domains(), None)
        privileged.set_domain(domain)

    return domain


def get_domains():
    """Return a list with the configured domains."""
    if not pathlib.Path('/var/lib/mumble-server/').exists():
        return []

    domain = get_domain()
    if domain:
        return [domain]

    return []


def _diagnose_config() -> DiagnosticCheck:
    """Check that configuration changes made during setup are in place."""
    result = Result.PASSED if privileged.check_setup() else Result.FAILED
    return DiagnosticCheck('mumble-config',
                           gettext_noop('Mumble server is configured'), result)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configuration helper for bepasty."""

import collections
import json
import pathlib
import secrets
import shutil
import string

import augeas

from plinth import action_utils
from plinth.actions import privileged, secret_str
from plinth.modules import bepasty

CONF_FILE = pathlib.Path('/etc/bepasty-freedombox.conf')
DATA_DIR = '/var/lib/private/bepasty'
PASSWORD_LENGTH = 20
SERVICE_NAME = 'uwsgi-app@bepasty-freedombox.service'


def _augeas_load():
    """Initialize Augeas."""
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.set('/augeas/load/Simplevars/lens', 'Simplevars.lns')
    aug.set('/augeas/load/Simplevars/incl[last() + 1]', str(CONF_FILE))
    aug.load()
    return aug


def _key_path(key):
    """Return the augeas path for the key."""
    return '/files' + str(CONF_FILE) + '/' + key


def conf_file_read():
    """Read and return the configuration."""
    aug = _augeas_load()
    conf = collections.OrderedDict()
    for path in aug.match(_key_path('*')):
        key = path.rsplit('/', 1)[-1]
        if key[0] != '#':
            conf[key] = json.loads(aug.get(path))

    return conf


def conf_file_write(conf):
    """Write configuration to the file."""
    aug = _augeas_load()
    for key, value in conf.items():
        if not key.startswith('#'):
            value = json.dumps(value)

        aug.set(_key_path(key), value)

    aug.save()


@privileged
def setup(domain_name: str):
    """Post installation actions for bepasty."""
    # Create configuration file if needed.
    if not CONF_FILE.is_file():
        passwords = [_generate_password() for _ in range(3)]
        conf = {
            '#comment':
                'This file is managed by FreedomBox. Only a small subset of '
                'the original configuration format is supported. Each line '
                'should be in KEY = VALUE format. VALUE must be a JSON '
                'encoded string.',
            'SITENAME': domain_name,
            'STORAGE_FILESYSTEM_DIRECTORY': '/var/lib/bepasty',
            'SECRET_KEY': secrets.token_hex(64),
            'PERMISSIONS': {
                passwords[0]: 'admin,list,create,read,delete',
                passwords[1]: 'list,create,read,delete',
                passwords[2]: 'list,read',
            },
            'PERMISSION_COMMENTS': {
                passwords[0]: 'admin',
                passwords[1]: 'editor',
                passwords[2]: 'viewer',
            },
            'DEFAULT_PERMISSIONS': 'read',
        }
        conf_file_write(conf)
        CONF_FILE.chmod(0o640)

    # Migrate from old bepasty:bepasty ownership to root:root
    shutil.chown(CONF_FILE, user='root', group='root')


@privileged
def get_configuration() -> dict[str, object]:
    """Get default permissions, passwords, permissions and comments."""
    return conf_file_read()


@privileged
def add_password(permissions: list[str], comment: str | None = None):
    """Generate a password with given permissions."""
    conf = conf_file_read()
    permissions = _format_permissions(permissions)
    password = _generate_password()
    conf['PERMISSIONS'][password] = permissions
    if comment:
        conf['PERMISSION_COMMENTS'][password] = comment

    conf_file_write(conf)
    # Service is started again by socket.
    action_utils.service_stop(SERVICE_NAME)


@privileged
def remove_password(password: secret_str):
    """Remove a password and its permissions."""
    conf = conf_file_read()
    if password in conf['PERMISSIONS']:
        del conf['PERMISSIONS'][password]

    if password in conf['PERMISSION_COMMENTS']:
        del conf['PERMISSION_COMMENTS'][password]
    conf_file_write(conf)
    action_utils.service_stop(SERVICE_NAME)


@privileged
def set_default(permissions: list[str]):
    """Set default permissions."""
    conf = {'DEFAULT_PERMISSIONS': _format_permissions(permissions)}
    conf_file_write(conf)
    action_utils.service_stop(SERVICE_NAME)


def _format_permissions(permissions=None):
    """Format permissions as comma-separated."""
    return ','.join(set(bepasty.PERMISSIONS.keys()).intersection(
        permissions)) if permissions else ''


def _generate_password():
    """Generate a random password."""
    alphabet = string.ascii_letters + string.digits
    return ''.join(secrets.choice(alphabet) for _ in range(PASSWORD_LENGTH))


@privileged
def uninstall():
    """Remove data and configuration file."""
    shutil.rmtree(DATA_DIR, ignore_errors=True)
    CONF_FILE.unlink(missing_ok=True)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app for Apache server.

This module implements a mechanism for protecting various URL paths using
FreedomBox's OpenID Connect implementation as Identity Provider and
mod_auth_oidc as Relying Party. The following is a simplified description of
how the flow works:

- User accesses the URL /foo using a browser. /foo is a URL path which is
  protected by this module's OpenID Connect SSO (using AuthType
  openid-connect).

- mod_auth_opendic seizes control and checks for authorization. Since this is
  the first visit, it starts the authentication/authorization process. It first
  redirects the browser to provider discovery URL
  /freedombox/apache/discover-idp/.

- This URL selects and Identity Provider based on incoming URL's host header.
  It will select https://mydomain.example/freedombox/o as IDP if the original
  URL is https://mydomain.example/foo. Or https://freedombox.local/freedombox/o
  if the original URL is https://freedombox.local/foo. After selection it will
  redirect the browser back to /apache/oidc/callback with the selected IDP in
  the GET parameters.

- /apache/oidc/callback is controlled by mod_auth_openidc which receives the
  IDP selection. It will then query the IDP for further information such as
  authorization URL, token URL, supported scopes and claims. This is done using
  a backend call to /freedombox/o/.well-known/openid-configuration.

- After determining the authorization end point (/freedombox/o/authorize/) from
  the metadata, mod_auth_openidc will start the authentication/authorization
  process by redirecting the browser to the URL.

- FreedomBox shows login page if the user is not already logged in. User logs
  in.

- FreedomBox will show a page asking the user to authorize the application to
  access information such as name and email. In case of Apache's
  mod_auth_openidc, this is skipped.

- FreedomBox will redirect back to /apache/oidc/callback after various checks.
  This request will contain authorization grant token and OIDC claims in
  parameters.

- mod_auth_openidc connects using back channel HTTP call to token endpoint
  (/freedombox/o/token/) with the authorization grant token and then obtains
  access token and refresh token. OIDC claims are checked using client_secret
  known only to FreedomBox IDP and mod_auth_openidc.

- The OIDC claims contains username as part of 'sub' claim. This is exported as
  REMOTE_USER header. 'freedombox_groups' contains the list of groups that
  FreedomBox account is part of. These, along with 'Require claim' Apache
  configuration directives, are used to determine if the user should get access
  to /foo path or not.

- The application providing /foo will have access to information such username
  and groups as part of REMOTE_USER and other OIDC_* environment variables.

- mod_auth_openidc also sets cookies that ensure that the whole process is not
  repeated when a second request for the path /foo is received.
"""

import ipaddress
import os

from django.utils.translation import gettext_lazy as _

from plinth import action_utils
from plinth import app as app_module
from plinth import cfg
from plinth.config import DropinConfigs
from plinth.daemon import Daemon, RelatedDaemon
from plinth.modules import names
from plinth.modules.firewall.components import Firewall
from plinth.modules.letsencrypt.components import LetsEncrypt
from plinth.modules.oidc.components import OpenIDConnect
from plinth.package import Packages
from plinth.signals import domain_added, domain_removed
from plinth.utils import format_lazy, is_valid_user_name

from . import privileged


class ApacheApp(app_module.App):
    """FreedomBox app for Apache web server."""

    app_id = 'apache'

    _version = 17

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               is_essential=True, name=_('Apache HTTP Server'))
        self.add(info)

        packages = Packages('packages-apache', [
            'apache2', 'php-fpm', 'ssl-cert', 'uwsgi-core',
            'uwsgi-plugin-python3', 'libapache2-mod-auth-openidc'
        ])
        self.add(packages)

        dropin_configs = DropinConfigs('dropin-configs-apache', [
            '/etc/apache2/conf-available/10-freedombox.conf',
            '/etc/apache2/conf-available/php-fpm-freedombox.conf',
            '/etc/fail2ban/jail.d/apache-auth-freedombox.conf',
        ])
        self.add(dropin_configs)

        web_server_ports = Firewall('firewall-web', _('Web Server'),
                                    ports=['http', 'https'], is_external=True)
        self.add(web_server_ports)

        freedombox_ports = Firewall(
            'firewall-plinth',
            format_lazy(_('{box_name} Web Interface (Plinth)'),
                        box_name=_(cfg.box_name)), ports=['http', 'https'],
            is_external=True)
        self.add(freedombox_ports)

        letsencrypt = LetsEncrypt('letsencrypt-apache', domains='*',
                                  daemons=['apache2'], reload_daemons=True)
        self.add(letsencrypt)

        openidconnect = OpenIDConnect(
            'openidconnect-apache', 'apache',
            _('Web app protected by FreedomBox'),
            redirect_uris=['https://{domain}/apache/oidc/callback'],
            skip_authorization=True)
        self.add(openidconnect)

        daemon = Daemon('daemon-apache', 'apache2')
        self.add(daemon)

        # To be able to disable the old uwsgi init.d script.
        related_daemon = RelatedDaemon('related-daemon-apache', 'uwsgi')
        self.add(related_daemon)

    @staticmethod
    def post_init():
        """Perform post initialization operations."""
        domain_added.connect(_on_domain_added)
        domain_removed.connect(_on_domain_removed)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup(old_version)
        self.enable()


def validate_host(hostname: str):
    """Check whether we are allowed to be called by a given name.

    This is to prevent DNS rebinding attacks and other poor consequences in the
    OpenID Connect protoctol.
    """
    if hostname in ('localhost', 'ip6-localhost', 'ip6-loopback'):
        return

    if hostname == action_utils.get_hostname():
        return

    if hostname in names.components.DomainName.list_names():
        return

    try:
        ipaddress.ip_address(hostname)
        return
    except ValueError:
        pass

    raise ValueError(f'Server not configured to be called as {hostname}')


def setup_oidc_client(netloc: str, hostname: str):
    """Setup OpenID Connect client configuration.

    netloc is hostname or IP address along with port as parsed by
    urllib.parse.urlparse() method from a URL.
    """
    validate_host(hostname)

    oidc = app_module.App.get('apache').get_component('openidconnect-apache')
    privileged.setup_oidc_client(netloc, oidc.client_id,
                                 oidc.get_client_secret())


def _on_domain_added(sender, domain_type, name='', description='',
                     services=None, **kwargs):
    """Add site specific configuration for a domain."""
    if name:
        privileged.domain_setup(name)


def _on_domain_removed(sender, domain_type, name='', **kwargs):
    """Remove site specific configuration for a domain."""
    if name:
        privileged.domain_remove(name)


# (U)ser (W)eb (S)ites


def uws_directory_of_user(user):
    """Return the directory of the given user's website."""
    return '/home/{}/public_html'.format(user)


def uws_url_of_user(user):
    """Return the url path of the given user's website."""
    return '/~{}/'.format(user)


def user_of_uws_directory(directory):
    """Return the user of a given user website directory."""
    if directory.startswith('/home/'):
        pos_ini = 6
    elif directory.startswith('home/'):
        pos_ini = 5
    else:
        return None

    pos_end = directory.find('/public_html')
    if pos_end == -1:
        return None

    user = directory[pos_ini:pos_end]
    return user if is_valid_user_name(user) else None


def user_of_uws_url(url):
    """Return the user of a given user website url path."""
    MISSING = -1

    pos_ini = url.find('~')
    if pos_ini == MISSING:
        return None

    pos_end = url.find('/', pos_ini)
    if pos_end == MISSING:
        pos_end = len(url)

    user = url[pos_ini + 1:pos_end]
    return user if is_valid_user_name(user) else None


def uws_directory_of_url(url):
    """Return the directory of the user's website for the given url path.

    Note: It doesn't return the full OS file path to the url path!
    """
    return uws_directory_of_user(user_of_uws_url(url))


def uws_url_of_directory(directory):
    """Return the url base path of the user's website for the given OS path.

    Note: It doesn't return the url path for the file!
    """
    return uws_url_of_user(user_of_uws_directory(directory))


def get_users_with_website():
    """Return a dictionary of users with actual website subdirectory."""

    def lst_sub_dirs(directory):
        """Return the list of subdirectories of the given directory."""
        return [
            name for name in os.listdir(directory)
            if os.path.isdir(os.path.join(directory, name))
        ]

    return {
        name: uws_url_of_user(name)
        for name in lst_sub_dirs('/home')
        if os.path.isdir(uws_directory_of_user(name))
    }



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Tests for forms in names app."""

from ..forms import DomainAddForm, HostnameForm


def test_hostname_field():
    """Test that hostname field accepts only valid hostnames."""
    valid_hostnames = [
        'a', '0a', 'a0', 'AAA', '00', '0-0', 'example-hostname', 'example',
        '012345678901234567890123456789012345678901234567890123456789012'
    ]
    invalid_hostnames = [
        '', '-', '-a', 'a-', '.a', 'a.', 'a.a', '?', 'a?a',
        '0123456789012345678901234567890123456789012345678901234567890123'
    ]

    for hostname in valid_hostnames:
        form = HostnameForm({'hostname': hostname})
        assert form.is_valid()

    for hostname in invalid_hostnames:
        form = HostnameForm({'hostname': hostname})
        assert not form.is_valid()


def test_domain_name_field():
    """Test that domain name field accepts only valid domain names."""
    valid_domain_names = [
        'a', '0a', 'a0', 'AAA', '00', '0-0', 'example-hostname', 'example',
        'example.org', 'a.b.c.d', 'a-0.b-0.c-0',
        '012345678901234567890123456789012345678901234567890123456789012',
        ((('x' * 63) + '.') * 3) + 'x' * 61
    ]
    invalid_domain_names = [
        '', '-', '-a', 'a-', '.a', 'a.', '?', 'a?a', 'a..a', 'a.-a', '.',
        ((('x' * 63) + '.') * 3) + 'x' * 62, 'x' * 64
    ]

    for domain_name in valid_domain_names:
        form = DomainAddForm({'domain_name': domain_name})
        assert form.is_valid()

    for domain_name in invalid_domain_names:
        form = DomainAddForm({'domain_name': domain_name})
        assert not form.is_valid()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox app for help pages.
"""

import logging
import os

from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext_lazy

from plinth import app as app_module
from plinth import cfg, menu, web_server

logger = logging.getLogger(__name__)


class HelpApp(app_module.App):
    """FreedomBox app for showing help."""

    app_id = 'help'

    _version = 1

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               is_essential=True)
        self.add(info)

        menu_item = menu.Menu('menu-help', _('Help'), 'fa-book', None,
                              'help:index', parent_url_name='index')
        self.add(menu_item)
        menu_item = menu.Menu('menu-help-manual',
                              pgettext_lazy('User guide', 'Manual'),
                              'fa-info-circle', None, 'help:manual',
                              parent_url_name='help:index', order=10)
        self.add(menu_item)
        menu_item = menu.Menu('menu-help-support', _('Get Support'),
                              'fa-life-ring', None, 'help:support',
                              parent_url_name='help:index', order=20)
        self.add(menu_item)
        menu_item = menu.Menu('menu-help-feedback', _('Submit Feedback'),
                              'fa-comments', None, 'help:feedback',
                              parent_url_name='help:index', order=25)
        self.add(menu_item)
        menu_item = menu.Menu('menu-help-contribute', _('Contribute'),
                              'fa-wrench', None, 'help:contribute',
                              parent_url_name='help:index', order=30)
        self.add(menu_item)
        menu_item = menu.Menu('menu-help-about', _('About'), 'fa-star', None,
                              'help:about', parent_url_name='help:index',
                              order=100)
        self.add(menu_item)

    def post_init(self):
        """Perform post initialization operations."""
        directory_map = {}
        try:
            langs = os.listdir(os.path.join(cfg.doc_dir, 'manual'))
            for lang in langs:
                manual_dir = os.path.join(cfg.doc_dir, 'manual', lang,
                                          'images')
                manual_url = f'/help/manual/{lang}/images'
                directory_map[manual_url] = manual_dir
        except FileNotFoundError:
            logger.error('Unable to find manual directory. Ensure that '
                         'freedombox-doc-* packages are installed')

        static_files = web_server.StaticFiles('static-files-help',
                                              directory_map)
        self.add(static_files)

        # Mounting has to be done manually because web server will been
        # initialized before post_init() runs. Web server initialization is
        # when all existing StaticFiles components are auto-mounted.
        static_files.mount()

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        self.enable()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Application manifest for bind.
"""

from django.utils.translation import gettext_lazy as _

backup = {
    'config': {
        'files': ['/etc/bind/named.conf.options']
    },
    'services': ['named']
}

tags = [
    _('DNS'),
    _('Server'),
    _('Resolver'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Django views for Searx."""

from django.contrib import messages
from django.utils.translation import gettext as _

from plinth import views
from plinth.modules import searx

from . import privileged
from .forms import SearxForm


class SearxAppView(views.AppView):
    """Serve configuration page."""

    app_id = 'searx'
    form_class = SearxForm

    def get_initial(self):
        """Return the status of the service to fill in the form."""
        initial = super().get_initial()
        initial['safe_search'] = privileged.get_safe_search()
        initial['public_access'] = searx.is_public_access_enabled()
        return initial

    def form_valid(self, form):
        """Apply the changes submitted in the form."""
        old_data = form.initial
        form_data = form.cleaned_data

        if str(old_data['safe_search']) != form_data['safe_search']:
            try:
                privileged.set_safe_search(int(form_data['safe_search']))
                messages.success(self.request, _('Configuration updated.'))
            except Exception:
                messages.error(self.request,
                               _('An error occurred during configuration.'))

        if old_data['public_access'] != form_data['public_access']:
            try:
                if form_data['public_access']:
                    searx.enable_public_access()
                else:
                    searx.disable_public_access()
                messages.success(self.request, _('Configuration updated.'))
            except Exception:
                messages.error(self.request,
                               _('An error occurred during configuration.'))

        return super().form_valid(form)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Setup Django web framework.
"""

import logging
import os
import pathlib
import random
import stat

import django.conf
import django.core.management
import django.core.wsgi
from django.conf import global_settings
from django.contrib.messages import constants as message_constants

from . import cfg, glib, log, module_loader, settings

logger = logging.getLogger(__name__)


def init():
    """Setup Django configuration in the absence of .settings file"""
    # Workaround for django-simple-captcha 0.5.6 not being compatible with
    # Django 3.2. 0.5.14 is almost there in Debian. Workaround only until then.
    django.utils.encoding.python_2_unicode_compatible = lambda x: x

    if cfg.secure_proxy_ssl_header:
        settings.SECURE_PROXY_SSL_HEADER = (cfg.secure_proxy_ssl_header,
                                            'https')

    if cfg.use_x_forwarded_for:
        settings.IPWARE_META_PRECEDENCE_ORDER = ('HTTP_X_FORWARDED_FOR', )

    settings.DATABASES['default']['NAME'] = cfg.store_file
    settings.DEBUG = cfg.develop
    settings.FORCE_SCRIPT_NAME = cfg.server_dir
    # Order our apps before django/library apps so that we can override their
    # templates, static files, management commands, etc. This requires that
    # proper dependencies are declared inside database migration scripts.
    settings.INSTALLED_APPS = module_loader.get_modules_to_load() + \
        settings.INSTALLED_APPS
    settings.LANGUAGES = get_languages()
    settings.LOGGING = log.get_configuration()
    settings.MESSAGE_TAGS = {message_constants.ERROR: 'danger'}
    settings.SECRET_KEY = _get_secret_key()
    settings.SESSION_FILE_PATH = os.path.join(cfg.data_dir, 'sessions')
    settings.STATIC_URL = '/'.join([cfg.server_dir,
                                    'static/']).replace('//', '/')
    settings.USE_X_FORWARDED_HOST = cfg.use_x_forwarded_host
    if cfg.develop:
        # Disable template caching in development so that page updates don't
        # require service restart.
        del settings.TEMPLATES[0]['APP_DIRS']
        settings.TEMPLATES[0]['OPTIONS']['loaders'] = [
            'django.template.loaders.filesystem.Loader',
            'django.template.loaders.app_directories.Loader',
        ]

    kwargs = {}
    for setting in dir(settings):
        if setting.isupper():
            kwargs[setting] = getattr(settings, setting)

    django.conf.settings.configure(**kwargs)
    django.setup(set_prefix=True)

    logger.debug('Configured Django with applications - %s',
                 ', '.join(settings.INSTALLED_APPS))


def post_init():
    """Perform operations after completing init of other modules."""
    logger.debug('Creating or adding new tables to data file')
    django.core.management.call_command('migrate', '--fake-initial',
                                        interactive=False, verbosity=0)
    os.chmod(cfg.store_file, stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP)

    # Cleanup expired sessions every day
    glib.schedule(24 * 3600, _cleanup_expired_sessions, in_thread=True)


def _get_secret_key():
    """Retrieve or create a new Django secret key."""
    secret_key_file = pathlib.Path(cfg.data_dir) / 'django-secret.key'
    if secret_key_file.exists():
        secret_key = secret_key_file.read_text()
        if len(secret_key) >= 128:
            return secret_key

    secret_key = _generate_secret_key()
    # File should be created with permission 0o700
    old_umask = os.umask(0o077)
    try:
        secret_key_file.write_text(secret_key)
    finally:
        os.umask(old_umask)

    return secret_key


def _generate_secret_key():
    """Generate a new random secret key for use with Django."""
    # We could have used django.core.management.utils.get_random_secret_key
    # but it is not documented and should be considered private.
    length = 128
    chars = 'abcdefghijklmnopqrstuvwxyz0123456789!@#$%^&*(-_=+)'
    rand = random.SystemRandom()
    return ''.join(rand.choice(chars) for _ in range(length))


def get_languages():
    """Return list of languages to show in the interface.

    Add additional languages that FreedomBox support but Django doesn't.

    """

    def gettext_noop(string):
        """Django's actual translation methods need Django to be setup."""
        return string

    return sorted(
        list(global_settings.LANGUAGES) + [
            ('gu', gettext_noop('Gujarati')),
        ])


def _cleanup_expired_sessions(data):
    """Cleanup expired Django sessions."""
    verbosity = 1 if cfg.develop else 0
    django.core.management.call_command('clearsessions', verbosity=verbosity)


def get_wsgi_application():
    """Return Django wsgi application."""
    return django.core.wsgi.get_wsgi_application()


def get_static_url():
    """Return Django static URL."""
    return django.conf.settings.STATIC_URL



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Forms for configuring unattended-upgrades.
"""

from django import forms
from django.utils.translation import gettext_lazy as _

from plinth.modules import upgrades


class ConfigureForm(forms.Form):
    """Configuration form to enable/disable automatic upgrades."""
    auto_upgrades_enabled = forms.BooleanField(
        label=_('Enable auto-update'), required=False, help_text=_(
            'When enabled, FreedomBox automatically updates once a day.'))

    dist_upgrade_enabled = forms.BooleanField(
        label=_('Enable auto-update to next stable release'), required=False,
        help_text=_('When enabled, FreedomBox will update to the next stable '
                    'distribution release when it is available.'))

    def __init__(self, *args, **kwargs):
        """Disable options as necessary."""
        super().__init__(*args, **kwargs)

        self.fields['dist_upgrade_enabled'].disabled = \
            not upgrades.can_enable_dist_upgrade()


class BackportsFirstbootForm(forms.Form):
    """Form to configure backports during first boot wizard."""
    backports_enabled = forms.BooleanField(
        label=_('Activate frequent feature updates (recommended)'),
        required=False, initial=True)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the minidlna Server module.
"""

from django.urls import re_path

from plinth.modules.minidlna.views import MiniDLNAAppView

urlpatterns = [
    re_path(r'^apps/minidlna/$', MiniDLNAAppView.as_view(), name='index'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""URLs for the GNOME module."""

from django.urls import re_path

from plinth.views import AppView

urlpatterns = [
    re_path(r'^apps/gnome/$', AppView.as_view(app_id='gnome'), name='index'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure Shadowsocks Server."""

import json
import os
import pathlib
import random
import string

from plinth import action_utils
from plinth.actions import privileged, secret_str

SHADOWSOCKS_CONFIG_SYMLINK = '/etc/shadowsocks-libev/fbxserver.json'
SHADOWSOCKS_CONFIG_ACTUAL = \
    '/var/lib/private/shadowsocks-libev/fbxserver/fbxserver.json'


@privileged
def setup():
    """Perform initial setup steps."""
    # Disable the default service, and use the templated service instead, so
    # that the configuration can be customized.
    action_utils.service_disable('shadowsocks-libev')

    os.makedirs('/var/lib/private/shadowsocks-libev/fbxserver/', exist_ok=True)

    if not os.path.islink(SHADOWSOCKS_CONFIG_SYMLINK):
        os.symlink(SHADOWSOCKS_CONFIG_ACTUAL, SHADOWSOCKS_CONFIG_SYMLINK)

    if not os.path.isfile(SHADOWSOCKS_CONFIG_ACTUAL):
        password = ''.join(
            random.choice(string.ascii_letters) for _ in range(12))
        initial_config = {
            'server': ['::0', '0.0.0.0'],  # As recommended in man page
            'mode': 'tcp_and_udp',
            'server_port': 8388,
            'password': password,
            'timeout': 86400,
            'method': 'chacha20-ietf-poly1305'
        }
        _merge_config(initial_config)

    from plinth.modules.shadowsocksserver import ShadowsocksServerApp
    if action_utils.service_is_enabled(ShadowsocksServerApp.DAEMON):
        action_utils.service_restart(ShadowsocksServerApp.DAEMON)


@privileged
def get_config() -> dict[str, int | str]:
    """Read and print Shadowsocks Server configuration."""
    config = open(SHADOWSOCKS_CONFIG_SYMLINK, 'r', encoding='utf-8').read()
    return json.loads(config)


def _merge_config(config):
    """Write merged configuration into file."""
    try:
        current_config = open(SHADOWSOCKS_CONFIG_SYMLINK, 'r',
                              encoding='utf-8').read()
        current_config = json.loads(current_config)
    except (OSError, json.JSONDecodeError):
        current_config = {}

    new_config = current_config
    new_config.update(config)
    new_config = json.dumps(new_config, indent=4, sort_keys=True)
    open(SHADOWSOCKS_CONFIG_SYMLINK, 'w', encoding='utf-8').write(new_config)


@privileged
def merge_config(password: secret_str, method: str):
    """Configure Shadowsocks Server."""
    config = {'password': password, 'method': method}
    _merge_config(config)

    # Don't try_restart because initial configuration may not be valid so
    # shadowsocks will not be running even when enabled.
    from . import ShadowsocksServerApp
    if action_utils.service_is_enabled(ShadowsocksServerApp.DAEMON):
        action_utils.service_restart(ShadowsocksServerApp.DAEMON)


@privileged
def uninstall():
    """Remove configuration files."""
    for path in SHADOWSOCKS_CONFIG_SYMLINK, SHADOWSOCKS_CONFIG_ACTUAL:
        pathlib.Path(path).unlink(missing_ok=True)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configuration helper for calibre."""

import pathlib
import shutil

from plinth import action_utils
from plinth.actions import privileged
from plinth.modules import calibre

LIBRARIES_PATH = pathlib.Path('/var/lib/calibre-server-freedombox/libraries')


@privileged
def list_libraries() -> list[str]:
    """Return the list of libraries setup."""
    libraries = []
    for library in LIBRARIES_PATH.glob('*/metadata.db'):
        libraries.append(str(library.parent.name))

    return libraries


@privileged
def create_library(name: str):
    """Create an empty library."""
    calibre.validate_library_name(name)
    library = LIBRARIES_PATH / name
    library.mkdir(mode=0o755)  # Raise exception if already exists
    action_utils.run(
        ['calibredb', '--with-library', library, 'list_categories'],
        check=False)

    # Force systemd StateDirectory= logic to assign proper ownership to the
    # DynamicUser=
    shutil.chown(LIBRARIES_PATH.parent, 'root', 'root')
    action_utils.service_try_restart(calibre.CalibreApp.DAEMON)


@privileged
def delete_library(name: str):
    """Delete a library and its contents."""
    calibre.validate_library_name(name)
    library = LIBRARIES_PATH / name
    shutil.rmtree(library)
    action_utils.service_try_restart(calibre.CalibreApp.DAEMON)


@privileged
def uninstall():
    """Remove all libraries during uninstall."""
    shutil.rmtree(LIBRARIES_PATH)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure network manager.

During initial setup, configure networking for all wired and wireless devices
by creating network manager connections.
"""

import collections
import itertools
import logging
import re
import subprocess

from plinth import action_utils
from plinth.actions import privileged


def _sort_interfaces(interfaces: list[str]) -> list[str]:
    """Sort interfaces in a well-defined way: eth0, eth1, eth2, ... eth10."""

    def key_func(interface):
        parts = re.findall(r'(\D*)(\d*)', interface)
        parts = [(string, int(number) if number else number)
                 for string, number in parts]
        return list(itertools.chain(parts))

    return sorted(interfaces, key=key_func)


def _get_interfaces() -> dict[str, list[str]]:
    """Return all network interfaces by their type."""
    output = action_utils.run(
        ['nmcli', '--terse', '--fields', 'type,device', 'device'],
        check=True).stdout
    interfaces = collections.defaultdict(list)
    for line in output.decode().splitlines():
        type_, _, interface = line.partition(':')
        interfaces[type_].append(interface)

    for type_ in interfaces:
        interfaces[type_] = _sort_interfaces(interfaces[type_])

    return interfaces


def _add_connection(connection_name: str, interface: str,
                    remaining_arguments: list[str]):
    """Add an Ethernet/Wi-Fi connection of type regular or shared."""
    output = action_utils.run(
        ['nmcli', '--terse', '--fields', 'name,device', 'con', 'show'],
        check=True).stdout
    lines = output.decode().splitlines()
    if f'{connection_name}:{interface}' in lines:
        logging.info('Connection %s already exists for device %s, not adding.',
                     connection_name, interface)
    else:
        action_utils.run([
            'nmcli', 'con', 'add', 'con-name', connection_name, 'ifname',
            interface
        ] + remaining_arguments, check=True)


def _activate_connection(connection_name: str):
    """Activate a network connection in a background process."""
    subprocess.Popen(['nohup', 'nmcli', 'con', 'up', connection_name],
                     stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)


def _configure_regular_interface(interface: str, zone: str):
    """Create a connection that is not a shared connection."""
    connection_name = 'FreedomBox WAN'
    properties = {'connection.autoconnect': 'TRUE', 'connection.zone': zone}

    # Create n-m connection for a regular interface
    _add_connection(connection_name, interface, ['type', 'ethernet'])
    _set_connection_properties(connection_name, properties)
    _activate_connection(connection_name)

    logging.info('Configured interface %s for %s use as %s.', interface, zone,
                 connection_name)


def _configure_shared_interface(interface: str):
    """Create a shared connection that has traffic forwarding enabled.

    Shared connection means:
    - Self-assign an address and network
    - Start and manage DNS server (dnsmasq)
    - Start and manage DHCP server (dnsmasq)
    - Register address with mDNS
    - Add firewall rules for NATing from this interface
    """
    connection_name = f'FreedomBox LAN {interface}'
    properties = {
        'connection.autoconnect': 'TRUE',
        'connection.zone': 'internal',
        'ipv4.method': 'shared'
    }

    # Create n-m connection for eth1
    _add_connection(connection_name, interface, ['type', 'ethernet'])
    _set_connection_properties(connection_name, properties)
    _activate_connection(connection_name)

    logging.info('Configured interface %s for shared use as %s.', interface,
                 connection_name)


def _set_connection_properties(connection_name: str, properties: dict[str,
                                                                      str]):
    """Configure property key/values on a connection."""
    for key, value in properties.items():
        action_utils.run(
            ['nmcli', 'con', 'modify', connection_name, key, value],
            check=True)


def _configure_wireless_interface(interface: str):
    """Configure a wireless access point."""
    connection_name = f'FreedomBox {interface}'
    ssid = f'FreedomBox{interface}'
    secret = 'freedombox123'
    properties = {
        'connection.autoconnect': 'TRUE',
        'connection.zone': 'internal',
        'ipv4.method': 'shared',
        'wifi.mode': 'ap',
        'wifi-sec.key-mgmt': 'wpa-psk',
        'wifi-sec.psk': secret
    }

    _add_connection(connection_name, interface, ['type', 'wifi', 'ssid', ssid])
    _set_connection_properties(connection_name, properties)
    _activate_connection(connection_name)

    logging.info('Configured interface %s for shared use as %s', interface,
                 connection_name)


def _multi_wired_setup(interfaces: list[str]):
    """Configure all Ethernet connections on a system with many of them."""
    _configure_regular_interface(interfaces[0], 'external')

    for interface in interfaces[1:]:
        _configure_shared_interface(interface)


def _one_wired_setup(interface: str, interfaces: dict[str, list[str]]):
    """Configure an Ethernet connection on a system with only one."""
    if not len(interfaces['wifi']):
        _configure_regular_interface(interface, 'internal')
    else:
        _configure_regular_interface(interface, 'external')


def _wireless_setup(interfaces: list[str]):
    """Configure all wireless access points."""
    for interface in interfaces:
        _configure_wireless_interface(interface)


@privileged
def setup():
    """Create network manager connections.

    For a user who installed using freedombox-setup Debian package, when
    FreedomBox Service (Plinth) is run for the first time, don't run network
    setup. This is ensured by checking for the file
    /var/lib/freedombox/is-freedombox-disk-image which will not exist.

    For a user who installed using FreedomBox disk image, when FreedomBox
    Service (Plinth) runs for the first time, setup process executes and
    triggers the script due networks module being an essential module.
    """
    if not action_utils.is_disk_image():
        logging.info(
            'Not a FreedomBox disk image. Skipping network configuration.')
        return

    logging.info('Setting up network configuration...')
    interfaces = _get_interfaces()

    if len(interfaces['ethernet']) == 0:
        logging.info('No wired interfaces detected.')
    elif len(interfaces['ethernet']) == 1:
        _one_wired_setup(interfaces['ethernet'][0], interfaces)
    else:
        _multi_wired_setup(interfaces['ethernet'])

    _wireless_setup(interfaces['wifi'])

    logging.info('Done setting up network configuration.')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Views for the Apache app."""

from urllib.parse import urlencode, urlparse

from django.http import (HttpResponseBadRequest, HttpResponseRedirect,
                         HttpResponseServerError)
from django.views import View

from . import setup_oidc_client, validate_host

# By default 'openid' scope already included by mod_auth_openidc
OIDC_SCOPES = 'email freedombox_groups'


class DiscoverIDPView(View):
    """A view called by auth_openidc Apache module to find the IDP.

    According to documentation for auth_openidc: an Issuer selection can be
    passed back to the callback URL as in:
    <callback-url>?iss=[${issuer}|${domain}|${e-mail-style-account-name}]
      [parameters][&login_hint=<login-hint>][&scopes=<scopes>]
      [&auth_request_params=<params>]

    where the <iss> parameter contains the URL-encoded issuer value of the
    selected Provider (or...), [parameters] contains the additional parameters
    that were passed in on the discovery request (e.g.
    target_link_uri=<url>&x_csrf=<x_csrf>&method=<method>&scopes=<scopes>)
    """

    def get(self, request):
        """Redirect back to auth_openidc module after selecting a IDP."""
        target_link_uri = request.GET.get('target_link_uri', '')
        method = request.GET.get('method', 'get')
        x_csrf = request.GET.get('x_csrf', '')
        oidc_callback = request.GET.get('oidc_callback')

        if method != 'get':
            return HttpResponseBadRequest(f'Cannot handle "{method}" method')

        oidc_callback_parts = urlparse(oidc_callback)
        request_host = request.get_host()
        if request_host != oidc_callback_parts.netloc:
            return HttpResponseBadRequest(
                f'Cannot redirect from {request_host} to a different host '
                f'{oidc_callback_parts.netloc}')

        try:
            validate_host(oidc_callback_parts.hostname)
        except ValueError:
            return HttpResponseBadRequest(
                f'Accessed using unknown domain {request_host}. Please add '
                'the domain to list of configured domains.')

        try:
            setup_oidc_client(oidc_callback_parts.netloc,
                              oidc_callback_parts.hostname)
        except ValueError:
            return HttpResponseServerError(
                f'Server not configured to called as {request_host}')

        url = '/apache/oidc/callback'
        params = {
            'iss': f'https://{request_host}/freedombox/o',
            'target_link_uri': target_link_uri,
            'method': method,
            'x_csrf': x_csrf,
            'scopes': OIDC_SCOPES,
        }
        params = urlencode(params)
        return HttpResponseRedirect(f'{url}?{params}')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to manage an email server."""

import logging

from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

import plinth.app
from plinth import cfg, frontpage, menu
from plinth.config import DropinConfigs
from plinth.daemon import Daemon
from plinth.modules.apache.components import Webserver
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import (Firewall,
                                                FirewallLocalProtection)
from plinth.modules.letsencrypt.components import LetsEncrypt
from plinth.package import Packages
from plinth.privileged import service as service_privileged
from plinth.signals import domain_added, domain_removed
from plinth.utils import format_lazy, gettext_noop

from . import aliases, dovecot, manifest, privileged

_description = [
    _('This is a complete email server solution using Postfix, Dovecot, '
      'and Rspamd. Postfix sends and receives emails. Dovecot allows '
      'email clients to access your mailbox using IMAP and POP3. Rspamd deals '
      'with spam.'),
    _('Email server currently does not work with many free domain services '
      'including those provided by the FreedomBox Foundation. Many ISPs '
      'also restrict outgoing email. Some lift the restriction after an '
      'explicit request. See manual page for more information.'),
    format_lazy(
        _('Each user on {box_name} gets an email address like '
          'user@mydomain.example. They will also receive mail from all '
          'addresses that look like user+foo@mydomain.example. Further, '
          'they can add aliases to their email address. Necessary aliases '
          'such as "postmaster" are automatically created pointing to the '
          'first admin user.'), box_name=_(cfg.box_name)),
    _('<a href="/freedombox/apps/roundcube/">Roundcube app</a> provides web '
      'interface for users to access email.'),
    _('During installation, any other email servers in the system will be '
      'uninstalled.')
]

logger = logging.getLogger(__name__)


class EmailApp(plinth.app.App):
    """FreedomBox app for an email server."""

    app_id = 'email'

    _version = 9

    def __init__(self) -> None:
        """Initialize the email app."""
        super().__init__()

        info = plinth.app.Info(app_id=self.app_id, version=self._version,
                               name=_('Postfix/Dovecot'),
                               icon_filename='email', description=_description,
                               manual_page='Email', clients=manifest.clients,
                               tags=manifest.tags,
                               donation_url='https://rspamd.com/support.html')
        self.add(info)

        menu_item = menu.Menu('menu-email', info.name, info.icon_filename,
                              info.tags, 'email:index', parent_url_name='apps')
        self.add(menu_item)

        shortcut = frontpage.Shortcut(
            'shortcut-email', info.name, icon=info.icon_filename,
            description=info.description, manual_page=info.manual_page,
            configure_url=reverse_lazy('email:index'), clients=info.clients,
            tags=info.tags, login_required=True)
        self.add(shortcut)

        tags = [gettext_noop('More emails'), gettext_noop('Same mailbox')]
        shortcut = frontpage.Shortcut('shortcut-email-aliases',
                                      _('My Email Aliases'),
                                      icon=info.icon_filename, tags=tags,
                                      url=reverse_lazy('email:aliases'),
                                      login_required=True)
        self.add(shortcut)

        # Other likely install conflicts have been discarded:
        # - msmtp, nullmailer, sendmail don't cause install faults.
        # - qmail and smail are missing in Bullseye (Not tested,
        #   but less likely due to that).
        packages = Packages(
            'packages-email', [
                'postfix', 'postfix-sqlite', 'dovecot-pop3d', 'dovecot-imapd',
                'dovecot-lmtpd', 'dovecot-managesieved', 'dovecot-ldap',
                'dovecot-fts-xapian', 'rspamd', 'redis-server', 'openssl'
            ], conflicts=['exim4-base', 'exim4-config', 'exim4-daemon-light'],
            conflicts_action=Packages.ConflictsAction.REMOVE,
            rerun_setup_on_upgrade=True)
        self.add(packages)

        dropin_configs = DropinConfigs('dropin-configs-email', [
            '/etc/apache2/conf-available/email-freedombox.conf',
            '/etc/fail2ban/jail.d/dovecot-freedombox.conf',
            '/etc/postfix/freedombox-aliases.cf',
            '/etc/rspamd/local.d/freedombox-logging.inc',
            '/etc/rspamd/local.d/freedombox-milter-headers.conf',
            '/etc/rspamd/local.d/freedombox-redis.conf',
            '/etc/rspamd/local.d/freedombox-dkim-signing.conf'
        ])
        self.add(dropin_configs)
        dropin_configs_sieve = DropinConfigs('dropin-configs-email-sieve', [
            '/etc/dovecot/freedombox-sieve/learn-ham.sieve',
            '/etc/dovecot/freedombox-sieve/learn-spam.sieve',
            '/etc/dovecot/freedombox-sieve-after/sort-spam.sieve'
        ])
        self.add(dropin_configs_sieve)
        dropin_configs_dovecot = DovecotDropinConfigs(
            'dropin-configs-email-dovecot', [
                '/etc/dovecot/conf.d/05-freedombox-passdb.conf',
                '/etc/dovecot/conf.d/05-freedombox-userdb.conf',
                '/etc/dovecot/conf.d/15-freedombox-auth.conf',
                '/etc/dovecot/conf.d/15-freedombox-mail.conf',
                '/etc/dovecot/conf.d/90-freedombox-imap.conf',
                '/etc/dovecot/conf.d/90-freedombox-lmtp.conf',
                '/etc/dovecot/conf.d/90-freedombox-mailboxes.conf',
                '/etc/dovecot/conf.d/90-freedombox-master.conf',
                '/etc/dovecot/conf.d/90-freedombox-tls.conf',
                '/etc/dovecot/conf.d/95-freedombox-sieve.conf',
                '/etc/dovecot/conf.d/95-freedombox-fts.conf',
                '/etc/dovecot/conf.d/freedombox-ldap.conf.ext'
            ])
        self.add(dropin_configs_dovecot)

        listen_ports = [(25, 'tcp4'), (25, 'tcp6'), (465, 'tcp4'),
                        (465, 'tcp6'), (587, 'tcp4'), (587, 'tcp6')]
        daemon = Daemon('daemon-email-postfix', 'postfix',
                        listen_ports=listen_ports)
        self.add(daemon)

        listen_ports = [(143, 'tcp4'), (143, 'tcp6'), (993, 'tcp4'),
                        (993, 'tcp6'), (110, 'tcp4'), (110, 'tcp6'),
                        (995, 'tcp4'), (995, 'tcp6'), (4190, 'tcp4'),
                        (4190, 'tcp6')]
        daemon = Daemon('daemon-email-dovecot', 'dovecot',
                        listen_ports=listen_ports)
        self.add(daemon)

        listen_ports = [(11332, 'tcp4'), (11332, 'tcp6'), (11333, 'tcp4'),
                        (11333, 'tcp6'), (11334, 'tcp4'), (11334, 'tcp6')]
        daemon = Daemon('daemon-email-rspamd', 'rspamd',
                        listen_ports=listen_ports)
        self.add(daemon)

        daemon = Daemon('daemon-email-redis', 'redis-server',
                        listen_ports=[(6379, 'tcp4'), (6379, 'tcp6')])
        self.add(daemon)

        port_names = [
            'smtp', 'smtps', 'smtp-submission', 'imaps', 'pop3s', 'managesieve'
        ]
        firewall = Firewall('firewall-email', info.name, ports=port_names,
                            is_external=True)
        self.add(firewall)

        firewall_local_protection = FirewallLocalProtection(
            'firewall-local-protection-email', ['11334'])
        self.add(firewall_local_protection)

        # /rspamd location
        webserver = Webserver(
            'webserver-email',  # unique id
            'email-freedombox',  # config file name
            urls=['https://{host}/rspamd'])
        self.add(webserver)

        # Let's Encrypt event hook
        letsencrypt = LetsEncrypt(
            'letsencrypt-email-postfix', domains='*', daemons=['postfix'],
            should_copy_certificates=True,
            private_key_path='/etc/postfix/letsencrypt/{domain}/chain.pem',
            certificate_path='/etc/postfix/letsencrypt/{domain}/chain.pem',
            user_owner='root', group_owner='root', managing_app='email')
        self.add(letsencrypt)

        letsencrypt = LetsEncrypt(
            'letsencrypt-email-dovecot', domains='*', daemons=['dovecot'],
            should_copy_certificates=True,
            private_key_path='/etc/dovecot/letsencrypt/{domain}/privkey.pem',
            certificate_path='/etc/dovecot/letsencrypt/{domain}/cert.pem',
            user_owner='root', group_owner='root', managing_app='email')
        self.add(letsencrypt)

        backup_restore = BackupRestore('backup-restore-email',
                                       **manifest.backup)
        self.add(backup_restore)

    @staticmethod
    def post_init():
        """Perform post initialization operations."""
        domain_added.connect(on_domain_added)
        domain_removed.connect(on_domain_removed)

    def setup(self, old_version):
        """Install and configure the app."""
        # Install
        super().setup(old_version)

        # Setup
        privileged.setup_home()
        self.get_component('letsencrypt-email-postfix').setup_certificates()
        self.get_component('letsencrypt-email-dovecot').setup_certificates()
        privileged.domain.set_all_domains()
        aliases.first_setup()
        privileged.setup_postfix()
        aliases.setup_common_aliases(_get_first_admin())

        # Enable drop-in configuration files component for sieve (temporarily)
        # to ensure that sievec can compile.
        self.get_component('dropin-configs-email-sieve').enable()
        self.get_component('dropin-configs-email-dovecot').enable()
        service_privileged.try_restart('dovecot')
        privileged.setup_spam()

        # Restart daemons
        if self.is_enabled():
            service_privileged.restart('postfix')
            service_privileged.restart('dovecot')
            service_privileged.restart('rspamd')

        # Expose to public internet
        if old_version == 0:
            self.enable()
        elif old_version < 5:
            privileged.fix_incorrect_key_ownership()
            service_privileged.try_restart('rspamd')


class DovecotDropinConfigs(DropinConfigs):
    """Configure dovecot based on its package version."""

    def get_target_path(self, path):
        """Return Path object for a target path."""
        version = '2.3'
        if dovecot.is_version_24():
            version = '2.4'

        target_path = super().get_target_path(path)
        target_path = target_path.parent / version / target_path.name
        return target_path


def _get_first_admin():
    """Return an admin user in the system or None if non exist."""
    from django.contrib.auth.models import User
    users = User.objects.filter(groups__name='admin')
    return users[0].username if users else None


def on_domain_added(sender, domain_type, name, description='', services=None,
                    **kwargs):
    """Handle addition of a new domain."""
    app = plinth.app.App.get('email')
    if app.needs_setup():
        return

    privileged.domain.set_all_domains()


def on_domain_removed(sender, domain_type, name='', **kwargs):
    """Handle removal of a domain."""
    app = plinth.app.App.get('email')
    if app.needs_setup():
        return

    privileged.domain.set_all_domains()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test module for utilities.
"""

import tempfile
from unittest.mock import MagicMock, Mock

import pytest
import ruamel.yaml
from django.test.client import RequestFactory
from ruamel.yaml.compat import StringIO

from plinth.utils import (SafeFormatter, YAMLFile, is_user_admin,
                          is_valid_user_name)


def test_is_valid_user_name():
    """Test valid user names in Debian."""
    f = is_valid_user_name
    assert not f('this_user_name_is_too_long_to_be_valid')
    assert not f('-invalid')
    assert not f('not\tvalid')
    assert not f('not\nvalid')
    assert not f('not valid')
    assert not f('not:valid')
    assert not f('not/valid')
    assert not f('not\\valid')
    assert f('€.;#@|')


class TestIsAdminUser:
    """Test class for is_user_admin utility."""

    @staticmethod
    @pytest.fixture(name='web_request')
    def fixture_web_request():
        """Setup each test case before execution."""
        web_request = RequestFactory().get('/freedombox/mockapp')
        web_request.user = Mock()
        web_request.session = MagicMock()
        return web_request

    @staticmethod
    def test_is_false_for_anonymous_user(web_request):
        """Test anonymous user is reported as non-admin."""
        web_request.user = Mock()
        web_request.user.is_authenticated = False
        assert not is_user_admin(web_request)
        assert not is_user_admin(web_request, cached=True)

    @staticmethod
    def test_values_for_authenticated_users(web_request):
        """Test correct return values for authenticated users."""
        web_request.user.groups.filter().exists = Mock(return_value=False)
        assert not is_user_admin(web_request)
        web_request.user.groups.filter().exists = Mock(return_value=True)
        assert is_user_admin(web_request)

    @staticmethod
    def test_caching_of_values(web_request):
        """Test that caching of values for authenticate users."""
        session_mock = MagicMock()
        session_dict = {}
        session_mock.__setitem__.side_effect = session_dict.__setitem__
        session_mock.__getitem__.side_effect = session_dict.__getitem__
        session_mock.__contains__.side_effect = session_dict.__contains__
        web_request.session = session_mock

        mock = Mock(return_value=False)
        web_request.user.groups.filter().exists = mock
        assert not is_user_admin(web_request)
        mock.assert_called_once_with()
        session_mock.__setitem__.assert_called_once_with(
            'cache_user_is_admin', False)

        mock = Mock(return_value=False)
        web_request.user.groups.filter().exists = mock
        assert not is_user_admin(web_request, cached=True)
        mock.assert_not_called()
        session_mock.__getitem__.assert_called_once_with('cache_user_is_admin')

        mock = Mock(return_value=False)
        web_request.user.groups.filter().exists = mock
        assert not is_user_admin(web_request, cached=False)
        mock.assert_called_once_with()
        session_mock.__getitem__.assert_called_once_with('cache_user_is_admin')


class TestYAMLFileUtil:
    """Check updating YAML files"""

    kv_pair = {'key': 'value'}

    yaml = ruamel.yaml.YAML()
    yaml.preserve_quotes = True  # type: ignore [assignment]

    def test_update_empty_yaml_file(self):
        """
        Update an empty YAML file with content.
        """
        test_file = tempfile.NamedTemporaryFile()
        conf = {'property1': self.kv_pair}

        with YAMLFile(test_file.name) as file_conf:
            for key in conf:
                file_conf[key] = conf[key]

        with open(test_file.name, 'r', encoding='utf-8') as retrieved_conf:
            buffer = StringIO()
            self.yaml.dump(conf, buffer)
            assert retrieved_conf.read() == buffer.getvalue()

    def test_update_non_empty_yaml_file(self):
        """
        Update a non-empty YAML file with modifications
        """
        test_file = tempfile.NamedTemporaryFile()

        with open(test_file.name, 'w', encoding='utf-8') as conf_file:
            self.yaml.dump({'property1': self.kv_pair}, conf_file)

        with YAMLFile(test_file.name) as file_conf:
            file_conf['property2'] = self.kv_pair

        with open(test_file.name, 'r', encoding='utf-8') as retrieved_conf:
            file_conf = self.yaml.load(retrieved_conf)
            assert file_conf == {
                'property1': self.kv_pair,
                'property2': self.kv_pair
            }

    @staticmethod
    def test_context_exception():
        """Test that exception during update does not update file."""
        test_file = tempfile.NamedTemporaryFile()
        with pytest.raises(ValueError):
            with YAMLFile(test_file.name) as yaml_file:
                yaml_file['property1'] = 'value1'
                raise ValueError('Test')

        assert open(test_file.name, 'r', encoding='utf-8').read() == ''


@pytest.mark.parametrize('input_, output', (
    (('', [], {}), ''),
    (('{} {}', [10, 20], {}), '10 20'),
    (('{1} {0} {key1}', [10, 20], {
        'key1': 'value1'
    }), '20 10 value1'),
    (('{2} {1} {key1}', [10, 20], {}), '?2? 20 ?key1?'),
    (('{a[2]}', [], {
        'a': [1, 2, 3]
    }), '3'),
    (('{a[b]}', [], {
        'a': []
    }), '?a[b]?'),
    (('{a["b"]}', [], {}), '?a["b"]?'),
    (('{a.b}', [], {}), '?a.b?'),
))
def test_safe_string_formatter(input_, output):
    """Test the safe string formatter."""
    formatter = SafeFormatter()
    assert output == formatter.vformat(*input_)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for Shadowsocks Client app.
"""

import pytest

from plinth.tests import functional

pytestmark = [pytest.mark.apps, pytest.mark.shadowsocks]


class TestShadowsocksApp(functional.BaseAppTests):
    app_name = 'shadowsocks'
    has_service = True
    has_web = False

    def install_and_setup(self, session_browser):
        """Install the app and run setup."""
        super().install_and_setup(session_browser)
        _configure(session_browser, 'example.com', 'fakepassword')

    @pytest.mark.backups
    def test_backup_restore(self, session_browser):
        """Test backup and restore of configuration."""
        _configure(session_browser, 'example.com', 'beforebackup123')
        functional.backup_create(session_browser, 'shadowsocks',
                                 'test_shadowsocks')

        _configure(session_browser, 'example.org', 'afterbackup123')
        functional.backup_restore(session_browser, 'shadowsocks',
                                  'test_shadowsocks')

        assert functional.service_is_running(session_browser, 'shadowsocks')
        assert _get_configuration(session_browser) == ('example.com',
                                                       'beforebackup123')


def _configure(browser, server, password):
    """Configure shadowsocks client with given server details."""
    functional.visit(browser, '/freedombox/apps/shadowsocks/')
    browser.find_by_id('id_server').fill(server)
    browser.find_by_id('id_password').fill(password)
    functional.submit(browser, form_class='form-configuration')


def _get_configuration(browser):
    """Return the server and password currently configured in shadowsocks."""
    functional.visit(browser, '/freedombox/apps/shadowsocks/')
    server = browser.find_by_id('id_server').value
    password = browser.find_by_id('id_password').value
    return server, password



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure TiddlyWiki."""

import pathlib
import re
import shutil
import urllib.request

from plinth import action_utils
from plinth.actions import privileged

EMPTY_WIKI_FILE = 'https://ftp.freedombox.org/pub/tiddlywiki/empty.html'

wiki_dir = pathlib.Path('/var/lib/tiddlywiki')


def _set_ownership(path: pathlib.Path):
    """Makes www-data:www-data the owner of the give path."""
    shutil.chown(path, user='www-data', group='www-data')


@privileged
def setup():
    """Setup wiki dir and CGI script."""
    wiki_dir.mkdir(parents=True, exist_ok=True)
    _set_ownership(wiki_dir)


def _normalize_wiki_file_name(name):
    """Return a normalized file name from a wiki name."""
    file_name = name.replace(' ', '_')
    invalid_characters = r'[\/\\\:\*\?\"\'\<\>\|]'
    file_name = re.sub(invalid_characters, '', file_name)
    if not file_name.endswith('.html'):
        return file_name + '.html'

    return file_name


@privileged
def create_wiki(file_name: str):
    """Initialize wiki with the latest version of TiddlyWiki."""
    file_name = _normalize_wiki_file_name(file_name)
    response = urllib.request.urlopen(EMPTY_WIKI_FILE)
    file_path = wiki_dir / file_name
    if file_path.exists():
        raise ValueError('Wiki exists')

    file_path.write_bytes(response.read())
    _set_ownership(file_path)


@privileged
def add_wiki_file(file_name: str, temporary_file_path: str):
    """Add an uploaded wiki file."""
    action_utils.move_uploaded_file(temporary_file_path, wiki_dir, file_name,
                                    allow_overwrite=False, user='www-data',
                                    group='www-data', permissions=0o644)


@privileged
def rename_wiki(old_name: str, new_name: str):
    """Rename wiki file."""
    old_name = _normalize_wiki_file_name(old_name)
    new_name = _normalize_wiki_file_name(new_name)
    file_path = wiki_dir / new_name
    if file_path.exists():
        raise ValueError('Wiki exists')

    (wiki_dir / old_name).rename(file_path)


@privileged
def delete_wiki(file_name: str):
    """Delete one wiki file by name."""
    file_name = _normalize_wiki_file_name(file_name)
    (wiki_dir / file_name).unlink(missing_ok=True)


@privileged
def uninstall():
    """Delete all the wiki content."""
    shutil.rmtree(wiki_dir)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Setup CherryPy web server.
"""

import logging
import os
import pathlib
import sys
import warnings
from typing import ClassVar

import cherrypy

from . import app as app_module
from . import cfg, log, web_framework

logger = logging.getLogger(__name__)

# When an app installs a python module as a dependency and imports it. CherryPy
# will start monitoring it for changes. When the app is uninstalled, the module
# is removed from the system leading to change detected by CherryPy. The entire
# service is then restarted if it is in development mode. This could cause a
# temporary failure in requests served leading to failures in functional tests.
# Workaround this by preventing auto-reloading for all but FreedomBox's python
# modules.
AUTORELOAD_REGEX = r'^plinth'

_CUSTOM_STATIC_URL = '/custom/static'

_USER_CSS_PATH = 'css/user.css'


def get_custom_static_url():
    """Return the URL path fragment for custom static URL."""
    return f'{cfg.server_dir}{_CUSTOM_STATIC_URL}'


def get_user_css():
    """Return the URL path fragement for user CSS if it exists else None."""
    user_css_path = pathlib.Path(cfg.custom_static_dir) / _USER_CSS_PATH
    if not user_css_path.exists():
        return None

    return get_custom_static_url() + '/' + _USER_CSS_PATH


def _mount_static_directory(static_dir, static_url):
    config = {
        '/': {
            'tools.staticdir.root': static_dir,
            'tools.staticdir.on': True,
            'tools.staticdir.dir': '.'
        }
    }
    app = cherrypy.tree.mount(None, static_url, config)
    log.setup_cherrypy_static_directory(app)


def init():
    """Setup CherryPy server"""
    logger.info('Setting up CherryPy server')

    # Configure default server
    cherrypy.config.update({
        'server.max_request_body_size': 0,
        'server.socket_host': cfg.host,
        'server.socket_port': cfg.port,
        'server.thread_pool': 10,
        # Avoid stating files once per second in production
        'engine.autoreload.on': cfg.develop,
        'engine.autoreload.match': AUTORELOAD_REGEX,
    })

    def _logging_middleware(application):
        """A WSGI middleware to log messages before executing them."""

        def _wrapper(environ, start_response):
            """Log request, then hand control to original app."""
            logger.info("%s %s", environ['REQUEST_METHOD'],
                        environ['PATH_INFO'])
            return application(environ, start_response)

        return _wrapper

    application = web_framework.get_wsgi_application()
    cherrypy.tree.graft(_logging_middleware(application), cfg.server_dir)

    static_dir = os.path.join(cfg.file_root, 'static')
    _mount_static_directory(static_dir, web_framework.get_static_url())

    custom_static_dir = cfg.custom_static_dir
    custom_static_url = get_custom_static_url()
    if os.path.exists(custom_static_dir):
        _mount_static_directory(custom_static_dir, custom_static_url)
    else:
        logger.debug(
            'Not serving custom static directory %s on %s, '
            'directory does not exist', custom_static_dir, custom_static_url)

    _mount_static_directory('/usr/share/javascript', '/javascript')

    for app in app_module.App.list():
        module = sys.modules[app.__module__]
        module_path = os.path.dirname(module.__file__)
        static_dir = os.path.join(module_path, 'static')
        if not os.path.isdir(static_dir):
            continue

        urlprefix = "%s%s" % (web_framework.get_static_url(), app.app_id)
        _mount_static_directory(static_dir, urlprefix)

    for component in StaticFiles.list():
        component.mount()

    cherrypy.engine.signal_handler.subscribe()


def run(on_web_server_stop):
    """Start the web server and block it until exit."""
    with warnings.catch_warnings():
        # Suppress warning that some of the static directories don't exist.
        # Since there is no way to add/remove those tree mounts at will, we
        # need to add them all before hand even if they don't exist now. If the
        # directories becomes available later, CherryPy serves them just fine.
        warnings.filterwarnings(
            'ignore', '(.|\n)*is not an existing filesystem path(.|\n)*',
            UserWarning)
        cherrypy.engine.start()

    cherrypy.engine.subscribe('stop', on_web_server_stop)


def block():
    """Block the calling thread until web server exits."""
    cherrypy.engine.block()


class StaticFiles(app_module.FollowerComponent):
    """Component to serve static files shipped with an app.

    Any files in <app>/static directory will be automatically served on
    /static/<app>/ directory by FreedomBox. This allows each app to ship custom
    static files that are served by the web server.

    However, in some rare circumstances, a system folder will need to be served
    on a path for the app to work. This component allows declaring such
    directories and the web paths they should be served on.

    """

    _all_instances: ClassVar[dict[str, 'StaticFiles']] = {}

    def __init__(self, component_id, directory_map=None):
        """Initialize the component.

        component_id should be a unique ID across all components of an app and
        across all components.

        directory_map should be a dictionary with keys to be web paths and
        values to be absolute path of the directory on disk to serve. The
        static files from the directory are served over the given web path. The
        web path will be prepended with the FreedomBox's configured base web
        path. For example, {'/foo': '/usr/share/foo'} means that
        '/usr/share/foo/bar.png' will be served over '/freedombox/foo/bar.png'
        if FreedomBox is configured to be served on '/freedombox'.

        """
        super().__init__(component_id)
        self.directory_map = directory_map
        self._all_instances[component_id] = self

    @classmethod
    def list(cls):
        """Return a list of all instances."""
        return cls._all_instances.values()

    def mount(self):
        """Perform configuration of the web server to handle static files.

        Called by web server abstraction layer after web server has been setup.

        """
        if self.directory_map:
            for web_path, file_path in self.directory_map.items():
                web_path = '%s%s' % (cfg.server_dir, web_path)
                _mount_static_directory(file_path, web_path)


def resolve_static_path(url: str) -> pathlib.Path:
    """Convert a URL for static file into a file path."""
    url_parts = url.split('/')
    for app in app_module.App.list():
        if url_parts[0] != app.app_id:
            continue

        try:
            module = sys.modules[app.__module__]
        except KeyError:
            raise ValueError('Module for app not loaded')

        if not hasattr(module, '__file__') or not module.__file__:
            raise ValueError('Module file for app could not be found')

        module_path = pathlib.Path(module.__file__).parent
        static_dir = module_path / 'static'
        if not static_dir.is_dir():
            raise ValueError(
                f'No static directory available for app {app.app_id}')

        return static_dir / '/'.join(url_parts[1:])

    return pathlib.Path(cfg.file_root) / 'static' / url



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Application manfiest for letsencrypt.
"""

from django.utils.translation import gettext_lazy as _

# XXX: Backup and restore the Apache site configuration.
backup = {'secrets': {'directories': ['/etc/letsencrypt/']}}

tags = [_('Certificates'), _('HTTPS'), _('Web security'), _('Encryption')]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Django form for configuring Searx."""

from django import forms
from django.utils.translation import gettext_lazy as _


class SearxForm(forms.Form):
    """Searx configuration form."""

    safe_search = forms.ChoiceField(
        label=_('Safe Search'), help_text=_(
            'Select the default family filter to apply to your search results.'
        ), choices=((0, _('None')), (1, _('Moderate')), (2, _('Strict'))))

    public_access = forms.BooleanField(
        label=_('Allow Public Access'), help_text=_(
            'Allow this application to be used by anyone who can reach it.'),
        required=False)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the dynamicdns module
"""

from django.urls import re_path

from . import views

urlpatterns = [
    re_path(r'^sys/dynamicdns/$', views.DynamicDNSAppView.as_view(),
            name='index'),
    re_path(r'^sys/dynamicdns/domain/add/$', views.DomainView.as_view(),
            name='domain-add'),
    re_path(r'^sys/dynamicdns/domain/(?P<domain>[^/]+)/edit/$',
            views.DomainView.as_view(), name='domain-edit'),
    re_path(r'^sys/dynamicdns/domain/(?P<domain>[^/]+)/delete/$',
            views.DomainDeleteView.as_view(), name='domain-delete'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Utilities for regular updates and dist-upgrades."""

import pathlib
import re

import augeas

from plinth import action_utils
from plinth.modules.apache.components import check_url

RELEASE_FILE_URL = \
    'https://deb.debian.org/debian/dists/{}/Release'

DIST_UPGRADE_REQUIRED_FREE_SPACE = 5000000

sources_list = pathlib.Path('/etc/apt/sources.list')


def check_auto() -> bool:
    """Return whether automatic updates are enabled."""
    arguments = [
        'apt-config', 'shell', 'UpdateInterval',
        'APT::Periodic::Update-Package-Lists'
    ]
    output = action_utils.run(arguments, check=True).stdout.decode()
    update_interval = 0
    match = re.match(r"UpdateInterval='(.*)'", output)
    if match:
        update_interval = int(match.group(1))

    return bool(update_interval)


def get_http_protocol() -> str:
    """Return the protocol to use for newly added repository sources."""
    try:
        from plinth.modules.torproxy import utils
        if utils.is_apt_transport_tor_enabled():
            return 'tor+http'
    except Exception:
        pass

    return 'http'


def is_release_file_available(protocol: str, dist: str,
                              backports=False) -> bool:
    """Return whether the release for dist[-backports] is available."""
    wrapper = None
    if protocol == 'tor+http':
        wrapper = 'torsocks'

    if backports:
        dist += '-backports'

    try:
        return check_url(RELEASE_FILE_URL.format(dist), wrapper=wrapper)
    except FileNotFoundError:
        return False


def is_sufficient_free_space() -> bool:
    """Return whether there is sufficient free space for dist upgrade."""
    output = action_utils.run(['df', '--output=avail', '/'], check=True).stdout
    free_space = int(output.decode().split('\n')[1])
    return free_space >= DIST_UPGRADE_REQUIRED_FREE_SPACE


def get_sources_list_codename() -> str | None:
    """Return the codename set in the /etc/apt/sources.list file."""
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.transform('aptsources', str(sources_list))
    aug.set('/augeas/context', '/files' + str(sources_list))
    aug.load()

    dists = set()
    for match_ in aug.match('*'):
        dist = aug.get(match_ + '/distribution')
        if not dist:
            continue

        dist = dist.removesuffix('-updates')
        dist = dist.removesuffix('-security')
        dists.add(dist)

    if 'unstable' in dists or 'sid' in dists:
        return 'unstable'

    if 'testing' in dists:
        return 'testing'

    # Multiple distributions are not understood.
    if len(dists) != 1:
        return None

    return dists.pop()


def get_current_release():
    """Return current release and codename as a tuple."""
    output = action_utils.run(
        ['lsb_release', '--release', '--codename', '--short'],
        check=True).stdout.decode().strip()
    lines = output.split('\n')
    return lines[0], lines[1]


def is_distribution_unstable() -> bool:
    """Return whether the current distribution is unstable.

    There is no way to distinguish between 'testing' and 'unstable'
    distributions in Debian using commands like lsb_release (powered by
    /etc/os-release). See: https://lwn.net/Articles/984635/ . So, use the value
    set in /etc/apt/sources.list.
    """
    codename = get_sources_list_codename()
    return codename in ['unstable', 'sid']


def is_distribution_rolling() -> bool:
    """Return whether the current distribution is testing or unstable."""
    # Release will be 'n/a' in latest unstable and testing distributions.
    release, _ = get_current_release()
    return release in ['unstable', 'testing', 'n/a']
